# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/9 15:42
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: peer_agent_template.py
from typing import Optional, Union

from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.message import Message
from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.agent.template.executing_agent_template import ExecutingAgentTemplate
from agentuniverse.agent.template.expressing_agent_template import ExpressingAgentTemplate
from agentuniverse.agent.template.planning_agent_template import PlanningAgentTemplate
from agentuniverse.agent.template.reviewing_agent_template import ReviewingAgentTemplate
from agentuniverse.agent.work_pattern.peer_work_pattern import PeerWorkPattern
from agentuniverse.agent.work_pattern.work_pattern_manager import WorkPatternManager
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger


class PeerAgentTemplate(AgentTemplate):
    planning_agent_name: str = "PlanningAgent"
    executing_agent_name: str = "ExecutingAgent"
    expressing_agent_name: str = "ExpressingAgent"
    reviewing_agent_name: str = "ReviewingAgent"
    eval_threshold: int = 60
    retry_count: int = 2
    jump_step: str = 'expressing'
    expert_framework: Optional[dict[str, Union[str, dict]]] = None

    def input_keys(self) -> list[str]:
        return ['input']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        agent_input.update({'eval_threshold': self.eval_threshold,
                            'retry_count': self.retry_count,
                            'jump_step': self.jump_step})
        return agent_input

    def execute(self, input_object: InputObject, agent_input: dict, **kwargs) -> dict:
        memory: Memory = self.process_memory(agent_input, **kwargs)
        agents = self._generate_agents()
        peer_work_pattern: PeerWorkPattern = WorkPatternManager().get_instance_obj('peer_work_pattern')
        peer_work_pattern = peer_work_pattern.set_by_agent_model(**agents)
        work_pattern_result = self.customized_execute(input_object=input_object, agent_input=agent_input, memory=memory,
                                                      peer_work_pattern=peer_work_pattern)
        self.add_peer_memory(memory, agent_input, work_pattern_result)
        return work_pattern_result

    async def async_execute(self, input_object: InputObject, agent_input: dict, **kwargs) -> dict:
        memory: Memory = self.process_memory(agent_input, **kwargs)
        agents = self._generate_agents()
        peer_work_pattern: PeerWorkPattern = WorkPatternManager().get_instance_obj('peer_work_pattern')
        peer_work_pattern = peer_work_pattern.set_by_agent_model(**agents)
        work_pattern_result = await self.customized_async_execute(input_object=input_object, agent_input=agent_input,
                                                                  memory=memory,
                                                                  peer_work_pattern=peer_work_pattern)
        self.add_peer_memory(memory, agent_input, work_pattern_result)
        return work_pattern_result

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory,
                           peer_work_pattern: PeerWorkPattern, **kwargs) -> dict:
        self.build_expert_framework(input_object)
        work_pattern_result = peer_work_pattern.invoke(input_object, agent_input)
        return work_pattern_result

    async def customized_async_execute(self, input_object: InputObject, agent_input: dict, memory: Memory,
                                       peer_work_pattern: PeerWorkPattern, **kwargs) -> dict:
        self.build_expert_framework(input_object)
        work_pattern_result = await peer_work_pattern.async_invoke(input_object, agent_input)
        return work_pattern_result

    def parse_result(self, agent_result: dict) -> dict:
        peer_results: list[dict] = agent_result.get('result', [])
        for item in reversed(peer_results):
            expressing_result = item.get('expressing_result')
            if expressing_result:
                return {'output': expressing_result.get('output')}

    def _generate_agents(self) -> dict:
        planning_agent = self._get_and_validate_agent(self.planning_agent_name, PlanningAgentTemplate)
        executing_agent = self._get_and_validate_agent(self.executing_agent_name, ExecutingAgentTemplate)
        expressing_agent = self._get_and_validate_agent(self.expressing_agent_name, ExpressingAgentTemplate)
        reviewing_agent = self._get_and_validate_agent(self.reviewing_agent_name, ReviewingAgentTemplate)
        return {'planning': planning_agent,
                'executing': executing_agent,
                'expressing': expressing_agent,
                'reviewing': reviewing_agent}

    @staticmethod
    def _get_and_validate_agent(agent_name: str, expected_type: type):
        agent = AgentManager().get_instance_obj(agent_name)
        if not agent:
            return None
        if not isinstance(agent, expected_type):
            raise ValueError(f"{agent_name} is not of the expected type {expected_type.__name__}")
        return agent

    def add_peer_memory(self, peer_memory: Memory, agent_input: dict, work_pattern_result: dict):
        if not peer_memory:
            return
        query = agent_input.get('input')
        message_list = []

        def _create_message_content(turn, role, agent_name, result):
            content = (f"Peer work pattern turn {turn + 1}: The agent responsible for {role} is: {agent_name}, "
                       f"Human: {query}, AI: {result}")
            return Message(source=agent_name, content=content)

        for i, single_turn_res in enumerate(work_pattern_result.get('result', [])):
            planning_result = single_turn_res.get('planning_result', {})
            if planning_result:
                message_list.append(_create_message_content(i, "planning and breaking down the task",
                                                            self.planning_agent_name, planning_result.get('framework')))

            executing_result = single_turn_res.get('executing_result', {})
            if executing_result:
                message_list.append(
                    _create_message_content(i, "executing the specific subtask", self.executing_agent_name,
                                            executing_result.get('executing_result')))

            expressing_result = single_turn_res.get('expressing_result', {})
            if expressing_result:
                message_list.append(
                    _create_message_content(i, "expressing and integrating the task into a final result",
                                            self.expressing_agent_name,
                                            expressing_result.get('output')))

            reviewing_result = single_turn_res.get('reviewing_result', {})
            if reviewing_result:
                message_list.append(_create_message_content(i, "reviewing and evaluating the result",
                                                            self.reviewing_agent_name,
                                                            reviewing_result.get('suggestion')))

        peer_memory.add(message_list, **agent_input)

    def build_expert_framework(self, input_object: InputObject):
        """Build expert framework from the expert framework tool selector or raw dictionary context.

        Args:
            input_object (InputObject): The input parameters passed by the user.

        Notes:
            The expert framework, whether using context or tool selector, must return a dictionary
             with keys for the specific content of planning, executing, expressing, and reviewing.
        """
        if self.expert_framework:
            context = self.expert_framework.get('context')
            selector = self.expert_framework.get('selector')
            # tool selector must return a dictionary
            # with keys for the specific content of planning, executing, expressing, and reviewing.
            if selector:
                selector_result = ToolManager().get_instance_obj(selector).run(**input_object.to_dict())
                if not isinstance(selector_result, dict):
                    raise ValueError("The expert framework tool selector must return a dictionary with keys"
                                     " for the specific content of planning, executing, expressing, and reviewing.")
                input_object.add_data('expert_framework', selector_result)
            elif context:  # raw dictionary context
                if not isinstance(context, dict):
                    raise ValueError("The expert framework raw context must be a dictionary with keys"
                                     " for the specific content of planning, executing, expressing, and reviewing.")
                input_object.add_data('expert_framework', context)

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'PeerAgentTemplate':
        super().initialize_by_component_configer(component_configer)
        planner_config = self.agent_model.plan.get('planner', {})
        if self.agent_model.profile.get('planning') is not None or planner_config.get('planning') is not None:
            self.planning_agent_name = self.agent_model.profile.get('planning') \
                if self.agent_model.profile.get('planning') is not None else planner_config.get('planning')
        if self.agent_model.profile.get('executing') is not None or planner_config.get('executing') is not None:
            self.executing_agent_name = self.agent_model.profile.get('executing') \
                if self.agent_model.profile.get('executing') is not None else planner_config.get('executing')
        if self.agent_model.profile.get('expressing') is not None or planner_config.get('expressing') is not None:
            self.expressing_agent_name = self.agent_model.profile.get('expressing') \
                if self.agent_model.profile.get('expressing') is not None else planner_config.get('expressing')
        if self.agent_model.profile.get('reviewing') is not None or planner_config.get('reviewing') is not None:
            self.reviewing_agent_name = self.agent_model.profile.get('reviewing') \
                if self.agent_model.profile.get('reviewing') is not None else planner_config.get('reviewing')
        if self.agent_model.profile.get('eval_threshold') or planner_config.get('eval_threshold'):
            self.eval_threshold = self.agent_model.profile.get('eval_threshold') or planner_config.get('eval_threshold')
        if self.agent_model.profile.get('retry_count') or planner_config.get('retry_count'):
            self.retry_count = self.agent_model.profile.get('retry_count') or planner_config.get('retry_count')
        if self.agent_model.profile.get('jump_step') or planner_config.get('jump_step'):
            self.jump_step = self.agent_model.profile.get('jump_step') or planner_config.get('jump_step')
        if self.agent_model.profile.get('expert_framework') or planner_config.get('expert_framework'):
            self.expert_framework = \
                self.agent_model.profile.get('expert_framework') or planner_config.get('expert_framework')
        self.memory_name = self.agent_model.memory.get('name')
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/17 17:33
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: prompt_manager.py
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase


@singleton
class PromptManager(ComponentManagerBase):
    """The PromptManager class, which is used to manage prompts."""

    def __init__(self):
        super().__init__(ComponentEnum.PROMPT)

    def get_instance_obj(self, component_instance_name: str, appname: str = None, new_instance: bool = False):
        return self._instance_obj_map.get(component_instance_name)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 14:31
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: thread_with_result.py

from threading import Thread

from agentuniverse.base.context.framework_context_manager import FrameworkContextManager


class ThreadWithReturnValue(Thread):
    """A thread can save the target func exec result."""

    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None):
        super().__init__(group, target, name, args, kwargs)

        if kwargs is None:
            kwargs = {}
        self.kwargs = kwargs
        self.args = args
        self.target = target
        self._return = None
        self.error = None
        self._context_values: dict = FrameworkContextManager().get_all_contexts()

    def run(self):
        """Run the target func and save result in _return."""
        if self.target is not None:
            context_tokens = {}
            # set the context values in the thread
            for var_name, var_value in self._context_values.items():
                token = FrameworkContextManager().set_context(var_name, var_value)
                context_tokens[var_name] = token

            try:
                self._return = self.target(*self.args, **self.kwargs)
            except Exception as e:
                self.error = e
            finally:
                if 'output_stream' in self.kwargs:
                    self.kwargs['output_stream'].put('{"type": "EOF"}')
                # finally, clear the context values in the thread
                for var_name, token in context_tokens.items():
                    FrameworkContextManager().reset_context(var_name, token)

    def result(self):
        """Wait for target func finished, then return the result or raise an
        error."""
        self.join()
        if self.error is not None:
            raise self.error
        return self._return



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/15 16:20
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: peer_planner.py
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent import Agent
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.base.util.logging.logging_util import LOGGER
from agentuniverse.base.util.agent_util import assemble_memory_output

default_sub_agents = {
    'planning': 'PlanningAgent',
    'executing': 'ExecutingAgent',
    'expressing': 'ExpressingAgent',
    'reviewing': 'ReviewingAgent',
}

default_jump_step = 'expressing'

default_eval_threshold = 60

default_retry_count = 2


class PeerPlanner(Planner):
    """Peer planner class."""

    def invoke(self, agent_model: AgentModel, planner_input: dict, input_object: InputObject) -> dict:
        """Invoke the planner.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): The input parameters passed by the user.
        Returns:
            dict: The planner result.
        """
        planner_config = agent_model.plan.get('planner')
        sub_agents = self.generate_sub_agents(planner_config)
        return self.agents_run(agent_model, sub_agents, planner_config, planner_input, input_object)

    @staticmethod
    def generate_sub_agents(planner_config: dict) -> dict:
        """Generate sub agents.

        Args:
            planner_config (dict): Planner config object.
        Returns:
            dict: Planner agents.
        """
        agents = dict()
        for config_key, default_agent in default_sub_agents.items():
            config_data = planner_config.get(config_key, None)
            if config_data == '':
                continue
            agents[config_key] = AgentManager().get_instance_obj(config_data if config_data else default_agent)
        return agents

    @staticmethod
    def build_expert_framework(planner_config: dict, input_object: InputObject):
        """Build expert framework for the given planner config object.

        Args:
            planner_config (dict): Planner config object.
            input_object (InputObject): Agent input object.
        """
        expert_framework = planner_config.get('expert_framework')
        if expert_framework:
            context = expert_framework.get('context')
            selector = expert_framework.get('selector')
            if selector:
                selector_result = ToolManager().get_instance_obj(selector).run(**input_object.to_dict())
                input_object.add_data('expert_framework', selector_result)
            elif context:
                input_object.add_data('expert_framework', context)

    def agents_run(self, agent_model: AgentModel, agents: dict, planner_config: dict, agent_input: dict,
                   input_object: InputObject) -> dict:
        """Planner agents run.

        Args:
            agent_model (AgentModel): Agent model object.
            agents (dict): Planner agents.
            planner_config (dict): Planner config object.
            agent_input (dict): Planner input object.
            input_object (InputObject): Agent input object.
        Returns:
            dict: The planner result.
        """
        result: dict = dict()

        loopResults = list()
        planning_result = dict()
        executing_result = dict()
        expressing_result = dict()
        reviewing_result = dict()

        retry_count = planner_config.get('retry_count', default_retry_count)
        jump_step = planner_config.get('jump_step', default_jump_step)
        eval_threshold = planner_config.get('eval_threshold', default_eval_threshold)

        self.build_expert_framework(planner_config, input_object)

        planningAgent: Agent = agents.get('planning')
        executingAgent: Agent = agents.get('executing')
        expressingAgent: Agent = agents.get('expressing')
        reviewingAgent: Agent = agents.get('reviewing')

        peer_memory: Memory = self.handle_memory(agent_model, agent_input)

        for _ in range(retry_count):
            LOGGER.info(f"Starting peer agents, retry_count is {_ + 1}.")
            if not planning_result or jump_step == "planning":
                planning_result = self.planning_agent_run(planningAgent, input_object, agent_input, peer_memory)

            if not executing_result or jump_step in ["planning", "executing"]:
                executing_result = self.executing_agent_run(executingAgent, input_object, agent_input, peer_memory)

            if not expressing_result or jump_step in ["planning", "executing", "expressing"]:
                expressing_result = self.expressing_agent_run(expressingAgent, input_object, agent_input, peer_memory)

            if not reviewing_result or jump_step in ["planning", "executing", "expressing", "reviewing"]:
                reviewing_result = self.reviewing_agent_run(reviewingAgent, input_object, agent_input, peer_memory)

            if not reviewing_result.to_dict() or (
                    reviewing_result.get_data('score') and reviewing_result.get_data('score') >= eval_threshold):
                loopResults.append({
                    "planning_result": planning_result,
                    "executing_result": executing_result,
                    "expressing_result": expressing_result,
                    "reviewing_result": reviewing_result
                })
                result['result'] = loopResults
                return result
            else:
                loopResults.append({
                    "planning_result": planning_result,
                    "executing_result": executing_result,
                    "expressing_result": expressing_result,
                    "reviewing_result": reviewing_result
                })
        result['result'] = loopResults
        return result

    def planning_agent_run(self, planning_agent: Agent, input_object: InputObject,
                           agent_input: dict, peer_memory: Memory) -> OutputObject:
        """Run the planning agent.

        Args:
            planning_agent (Agent): Planning agent object.
            input_object (InputObject): The input parameters passed by the user.
            agent_input (dict): Agent input object.
            peer_memory (Memory): Memory of the peer agent.
        Returns:
            OutputObject: The planning agent result.
        """
        planning_agent_name = planning_agent.agent_model.info.get('name') if planning_agent else ''
        if not planning_agent:
            LOGGER.warn("no planning agent.")
            planning_result = OutputObject({"framework": [agent_input.get('input')]})
        else:
            LOGGER.info(f"Starting planning agent.")
            planning_result = planning_agent.run(**input_object.to_dict())

        input_object.add_data('planning_result', planning_result)
        # add planning agent log info
        logger_info = f"\nPlanning agent execution result is :\n"
        for index, one_framework in enumerate(planning_result.get_data('framework')):
            logger_info += f"[{index + 1}] {one_framework} \n"
        LOGGER.info(logger_info)
        # add planning agent intermediate steps
        if planning_agent:
            self.stream_output(input_object, {"data": {
                'output': planning_result.get_data('framework'),
                "agent_info": planning_agent.agent_model.info
            }, "type": "planning"})
            content = (f"The agent responsible for planning and breaking down the task"
                       f" raised by users into several sub-tasks is: {planning_agent_name}, "
                       f"Human: {agent_input.get(self.input_key)}, "
                       f"AI: {planning_result.get_data('framework')}")
            assemble_memory_output(peer_memory, agent_input, content, planning_agent_name)
        return planning_result

    def executing_agent_run(self, executing_agent: Agent,
                            input_object: InputObject, agent_input: dict, peer_memory: Memory) -> OutputObject:
        """Run the executing agent.

        Args:
            executing_agent (Agent): Executing agent object.
            input_object (InputObject): The input parameters passed by the user.
            agent_input (dict): Agent input object.
            peer_memory (Memory): Memory of the peer agent.
        Returns:
            OutputObject: The executing agent result.
        """
        executing_agent_name = executing_agent.agent_model.info.get('name') if executing_agent else ''
        if not executing_agent:
            LOGGER.warn("no executing agent.")
            executing_result = OutputObject({})
        else:
            LOGGER.info(f"Starting executing agent.")
            executing_result = executing_agent.run(**input_object.to_dict())

        input_object.add_data('executing_result', executing_result)
        # add executing agent log info
        logger_info = f"\nExecuting agent execution result is :\n"
        if executing_result.get_data('executing_result'):
            for index, one_exec_res in enumerate(executing_result.get_data('executing_result')):
                one_exec_log_info = f"[{index + 1}] input: {one_exec_res['input']}\n"
                one_exec_log_info += f"[{index + 1}] output: {one_exec_res['output']}\n"
                logger_info += one_exec_log_info
        LOGGER.info(logger_info)
        # add executing agent intermediate steps
        if executing_agent:
            self.stream_output(input_object, {"data": {
                'output': executing_result.get_data('executing_result'),
                "agent_info": executing_agent.agent_model.info
            }, "type": "executing"})
            content = (
                f"The agent responsible for executing the specific subtask is: {executing_agent_name}, "
                f"Human: {agent_input.get(self.input_key)}, "
                f"AI: {executing_result.get_data('executing_result')}")
            assemble_memory_output(peer_memory, agent_input, content, executing_agent_name)
        return executing_result

    def expressing_agent_run(self, expressing_agent: Agent,
                             input_object: InputObject, agent_input: dict, peer_memory: Memory) -> OutputObject:
        """Run the expressing agent.

        Args:
            expressing_agent (Agent): Expressing agent object.
            input_object (InputObject): The input parameters passed by the user.
            agent_input (dict): Agent input object.
            peer_memory (Memory): Memory of the peer agent.
        Returns:
            OutputObject: The expressing agent result.
        """
        expressing_agent_name = expressing_agent.agent_model.info.get('name') if expressing_agent else ''
        if not expressing_agent:
            LOGGER.warn("no expressing agent.")
            expressing_result = OutputObject({})
        else:
            LOGGER.info(f"Starting expressing agent.")
            expressing_result = expressing_agent.run(**input_object.to_dict())

        input_object.add_data('expressing_result', expressing_result)
        # add expressing agent log info
        logger_info = f"\nExpressing agent execution result is :\n"
        logger_info += f"{expressing_result.get_data('output')}"
        LOGGER.info(logger_info)
        # add expressing agent intermediate steps
        if expressing_agent:
            self.stream_output(input_object, {"data": {
                'output': expressing_result.get_data('output'),
                "agent_info": expressing_agent.agent_model.info
            }, "type": "expressing"})
            content = (f"The agent responsible for expressing and integrating the task raised by user"
                       f" into a final result based on the results of several sub-tasks is: {expressing_agent_name}, "
                       f"Human: {agent_input.get(self.input_key)}, "
                       f"AI: {expressing_result.get_data('output')}")
            assemble_memory_output(peer_memory, agent_input, content, expressing_agent_name)

        return expressing_result

    def reviewing_agent_run(self, reviewing_agent: Agent,
                            input_object: InputObject, agent_input: dict, peer_memory: Memory) -> OutputObject:
        """Run the reviewing agent.

        Args:
            reviewing_agent (Agent): Reviewing agent object.
            input_object (InputObject): The input parameters passed by the user.
            agent_input (dict): Agent input object.
            peer_memory (Memory): Memory of the peer agent.
        Returns:
            OutputObject: The reviewing agent result.
        """
        reviewing_agent_name = reviewing_agent.agent_model.info.get('name') if reviewing_agent else ''
        if not reviewing_agent:
            LOGGER.warn("no reviewing agent.")
            reviewing_result = OutputObject({})
        else:
            LOGGER.info(f"Starting reviewing agent.")
            reviewing_result = reviewing_agent.run(**input_object.to_dict())

        input_object.add_data('reviewing_result', reviewing_result)
        # add reviewing agent log info
        logger_info = f"\nReviewing agent execution result is :\n"
        reviewing_info_str = f"review suggestion: {reviewing_result.get_data('suggestion')} \n"
        reviewing_info_str += f"review score: {reviewing_result.get_data('score')} \n"
        LOGGER.info(logger_info + reviewing_info_str)

        # add reviewing agent intermediate steps
        if reviewing_agent:
            # add reviewing agent intermediate steps
            self.stream_output(input_object,
                               {"data": {
                                   'output': reviewing_result.get_data('suggestion'),
                                   "agent_info": reviewing_agent.agent_model.info
                               }, "type": "reviewing"})
            content = (
                f"The agent responsible for reviewing"
                f" and evaluating the result to the task raised by user is {reviewing_agent_name}, "
                f"Human: {agent_input.get(self.input_key)}, "
                f"AI: {reviewing_result.get_data('suggestion')}")
            assemble_memory_output(peer_memory, agent_input, content, reviewing_agent_name)
        return reviewing_result



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 14:22
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_util.py
import copy
import os

from typing import Dict, List

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent import Agent
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.default.workflow_agent.workflow_agent import WorkflowAgent
from agentuniverse.agent.template.peer_agent_template import PeerAgentTemplate
from agentuniverse.agent.template.rag_agent_template import RagAgentTemplate
from agentuniverse.agent.template.react_agent_template import ReActAgentTemplate
from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.config.configer import Configer
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_configer import ProductConfiger
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.base.util.yaml_util import update_nested_yaml_value
from agentuniverse_product.service.model.agent_dto import AgentDTO
from agentuniverse_product.service.model.knowledge_dto import KnowledgeDTO
from agentuniverse_product.service.model.llm_dto import LlmDTO
from agentuniverse_product.service.model.planner_dto import PlannerDTO
from agentuniverse_product.service.model.prompt_dto import PromptDTO
from agentuniverse_product.service.model.tool_dto import ToolDTO
from agentuniverse_product.service.util.common_util import dict_does_not_contain_keys


def assemble_product_config_data(agent_dto: AgentDTO) -> Dict:
    """Assemble the agent product configuration data.

    Args:
        agent_dto (AgentDTO): The agent DTO.

    Returns:
        Dict: The assembled agent product configuration data.
    """
    return {
        'id': agent_dto.id,
        'nickname': agent_dto.nickname,
        'avatar': agent_dto.avatar,
        'opening_speech': agent_dto.opening_speech,
        'type': 'AGENT',
        'metadata': {
            'class': 'AgentProduct',
            'module': 'agentuniverse_product.base.agent_product',
            'type': 'PRODUCT'
        }
    }


def validate_create_agent_parameters(agent_dto: AgentDTO) -> None:
    """Validate the parameters for creating an agent.

    Args:
        agent_dto (AgentDTO): The agent DTO.
    """
    if agent_dto.id is None:
        raise ValueError("Agent id cannot be None.")
    agent = AgentManager().get_instance_obj(agent_dto.id)
    if agent:
        raise ValueError("Agent instance corresponding to the agent id already exists.")
    if agent_dto.planner is None:
        raise ValueError("The planner in agent cannot be None.")
    if agent_dto.planner.id != 'workflow_planner':
        if agent_dto.prompt is None:
            raise ValueError("The prompt in agent cannot be None.")
        if agent_dto.llm is None:
            raise ValueError("The llm in agent cannot be None.")


def assemble_agent_config_data(agent_dto: AgentDTO) -> Dict:
    """Assemble the agent configuration data.

    Args:
        agent_dto (AgentDTO): The agent DTO.

    Returns:
        Dict: The assembled agent configuration data.
    """
    agent_config_data = {
        'info': {
            'name': agent_dto.id,
            'description': agent_dto.description
        },
        'profile': {
        },
        'action': {}
    }
    if agent_dto.planner.workflow_id:
        agent_config_data['profile']['workflow_id'] = agent_dto.planner.workflow_id

    if agent_dto.llm:
        llm = LLMManager().get_instance_obj(agent_dto.llm.id)
        if llm is None:
            raise ValueError("The llm instance corresponding to the llm id cannot be found.")
        llm_model_dict = {'name': agent_dto.llm.id}
        if agent_dto.llm.temperature:
            llm_model_dict['temperature'] = agent_dto.llm.temperature
        if agent_dto.llm.model_name:
            llm_model_dict['model_name'] = agent_dto.llm.model_name[0]
        agent_config_data['profile']['llm_model'] = llm_model_dict

    if agent_dto.tool:
        agent_config_data['action']['tool'] = [tool.id for tool in agent_dto.tool]

    if agent_dto.knowledge:
        agent_config_data['action']['knowledge'] = [knowledge.id for knowledge in agent_dto.knowledge]

    if agent_dto.planner.id == 'workflow_planner':
        metadata_class = 'WorkflowAgent'
        metadata_agent_path = 'workflow_agent'
    elif agent_dto.planner.id == 'react_planner':
        metadata_class = 'ReActAgent'
        metadata_agent_path = 'react_agent'
        agent_config_data['profile']['prompt_version'] = 'default_react_agent.cn'
    else:
        metadata_class = 'RagAgent'
        metadata_agent_path = 'rag_agent'
        agent_config_data['profile']['prompt_version'] = 'default_rag_agent.cn'

    agent_config_data['metadata'] = {
        'class': metadata_class,
        'module': f'agentuniverse.agent.default.{metadata_agent_path}.{metadata_agent_path}',
        'type': 'AGENT'
    }
    return agent_config_data


def register_agent(file_path: str):
    """Register the agent instance to the agent manager.

    Args:
        file_path (str): The file path of the agent configuration.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    agent_configer: AgentConfiger = AgentConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(agent_configer)
    component_instance: Agent = component_clz().initialize_by_component_configer(agent_configer)
    component_instance.component_config_path = component_configer.configer.path
    AgentManager().register(component_instance.get_instance_code(), component_instance)


def register_product(file_path: str):
    """Register the product instance to the product manager.

    Args:
        file_path (str): The file path of the product configuration.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    product_configer: ProductConfiger = ProductConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(product_configer)
    component_instance: Product = component_clz().initialize_by_component_configer(product_configer)
    component_instance.component_config_path = component_configer.configer.path
    ProductManager().register(component_instance.get_instance_code(), component_instance)


def unregister_product(file_path: str):
    """Unregister the product instance from the product manager.

    Args:
        file_path (str): The file path of the product configuration.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    product_configer: ProductConfiger = ProductConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(product_configer)
    component_instance: Product = component_clz().initialize_by_component_configer(product_configer)
    ProductManager().unregister(component_instance.get_instance_code())


def assemble_agent_dto(agent: Agent, agent_dto: AgentDTO) -> AgentDTO:
    """Assemble agent dto from agent instance."""
    agent_model: AgentModel = agent.agent_model
    agent_dto.description = agent_model.info.get('description', '')
    agent_dto.prompt = get_prompt_dto(agent_model)
    agent_dto.llm = get_llm_dto(agent_model)
    agent_dto.memory = agent_model.memory.get('name', '')
    agent_dto.tool = get_tool_dto_list(agent_model)
    agent_dto.knowledge = get_knowledge_dto_list(agent_model)
    agent_dto.planner = get_planner_dto(agent)
    return agent_dto


def get_planner_dto(agent: Agent) -> PlannerDTO | None:
    """Get planner dto."""
    agent_model: AgentModel = agent.agent_model
    planner = agent_model.plan.get('planner', {})
    planner_name = planner.get('name') or get_default_planner_name(agent)
    if not planner_name:
        return None
    workflow_id = planner.get('workflow_id') if planner.get('workflow_id') else agent_model.profile.get('workflow_id')
    members = None
    if planner_name == 'peer_planner':
        members = assemble_peer_planner_members(planner, agent_model)
    return PlannerDTO(nickname='', id=planner_name, members=members,
                      workflow_id=workflow_id)


def assemble_peer_planner_members(planner: dict, agent_model: AgentModel) -> list[AgentDTO]:
    """Assemble members for 'peer_planner'."""
    member_keys = ['planning', 'executing', 'expressing', 'reviewing']
    default_member_names = ['PlanningAgent', 'ExecutingAgent', 'ExpressingAgent', 'ReviewingAgent']
    try:
        planner_data: dict = agent_model.profile if dict_does_not_contain_keys(planner, member_keys) else planner
        planner_data: dict = copy.deepcopy(planner_data)
        for i, member_key in enumerate(member_keys):
            planner_data.setdefault(member_key, default_member_names[i])
        return assemble_planner_members(planner_data, member_keys)
    except:
        return []


def get_default_planner_name(agent: Agent) -> str:
    """Return the default planner name based on the agent type."""
    if not agent:
        return ''
    if isinstance(agent, ReActAgentTemplate):
        return 'react_planner'
    elif isinstance(agent, RagAgentTemplate):
        return 'rag_planner'
    elif isinstance(agent, PeerAgentTemplate):
        return 'peer_planner'
    elif isinstance(agent, WorkflowAgent):
        return 'workflow_planner'
    return ''


def get_knowledge_dto_list(agent_model: AgentModel) -> List[KnowledgeDTO]:
    """Get knowledge dto list."""
    knowledge_name_list = agent_model.action.get('knowledge', [])
    res = []
    if not knowledge_name_list:
        return res
    for knowledge_name in knowledge_name_list:
        product: Product = ProductManager().get_instance_obj(knowledge_name)
        knowledge: Knowledge = KnowledgeManager().get_instance_obj(knowledge_name)
        if knowledge is None:
            continue
        knowledge_dto = KnowledgeDTO(nickname=product.nickname if product else '', id=knowledge_name)
        knowledge_dto.description = knowledge.description
        res.append(knowledge_dto)
    return res


def get_tool_dto_list(agent_model: AgentModel) -> List[ToolDTO]:
    """Get tool dto list."""
    tool_name_list = agent_model.action.get('tool', [])
    res = []
    if len(tool_name_list) < 1:
        return res
    for tool_name in tool_name_list:
        product: Product = ProductManager().get_instance_obj(tool_name)
        tool: Tool = ToolManager().get_instance_obj(tool_name)
        tool_dto = ToolDTO(nickname=product.nickname if product is not None else '',
                           avatar=product.avatar if product is not None else '',
                           id=tool.name)
        tool_dto.description = tool.description
        res.append(tool_dto)
    return res


def get_prompt_dto(agent_model: AgentModel) -> PromptDTO | None:
    """Get prompt dto."""
    profile_prompt_model: AgentPromptModel = AgentPromptModel(
        introduction=agent_model.profile.get('introduction'),
        target=agent_model.profile.get('target'),
        instruction=agent_model.profile.get('instruction'))

    prompt_version = agent_model.profile.get('prompt_version')
    version_prompt: Prompt = PromptManager().get_instance_obj(prompt_version)

    if version_prompt:
        version_prompt_model: AgentPromptModel = AgentPromptModel(
            introduction=getattr(version_prompt, 'introduction', ''),
            target=getattr(version_prompt, 'target', ''),
            instruction=getattr(version_prompt, 'instruction', ''))
        profile_prompt_model = profile_prompt_model + version_prompt_model
    if not profile_prompt_model:
        return None
    return PromptDTO(introduction=profile_prompt_model.introduction, target=profile_prompt_model.target,
                     instruction=profile_prompt_model.instruction)


def get_llm_dto(agent_model: AgentModel) -> LlmDTO | None:
    """Get llm dto from agent."""
    llm_model = agent_model.profile.get('llm_model', {})
    llm_id = llm_model.get('name')
    llm: LLM = LLMManager().get_instance_obj(llm_id)
    product: Product = ProductManager().get_instance_obj(llm_id)
    if llm is None:
        return None
    llm_model_name = llm_model.get('model_name') if llm_model.get('model_name') else llm.model_name
    llm_temperature = llm_model.get('temperature') if llm_model.get('temperature') else llm.temperature
    return LlmDTO(id=llm_id, nickname=product.nickname if product else '', temperature=llm_temperature,
                  model_name=[llm_model_name])


def assemble_planner_members(planner: dict, member_keys: List[str]) -> List[AgentDTO]:
    """Assemble planner members by looking up the configured multi-agent groups in the planner.

    Args:
        planner (dict): The planner dictionary in agent configuration.
        member_keys (List[str]): The multi-agent member keys to look up in the planner.

    Returns:
        List[AgentDTO]: The list of agent DTOs representing the planner members.
    """
    members = []
    if not planner or not member_keys:
        return members

    for member_key in member_keys:
        current_value = planner
        # get the corresponding agent instances configured in the planner through member_keys
        for path in member_key.split('.'):
            if isinstance(current_value, dict) and path in current_value:
                current_value = current_value[path]
            else:
                current_value = None
                break

        if not current_value:
            continue

        if isinstance(current_value, str):
            current_value = [current_value]

        if isinstance(current_value, list):
            for val in current_value:
                # assemble each agent instance
                agent: Agent = AgentManager().get_instance_obj(val)
                product: Product = ProductManager().get_instance_obj(val)
                if agent:
                    agent_dto = AgentDTO(id=val, nickname=product.nickname if product else '',
                                         avatar=product.avatar if product else '')
                    members.append(assemble_agent_dto(agent, agent_dto))

    return members


def update_agent_config(agent: Agent, agent_dto: AgentDTO, agent_config_path: str) -> None:
    """Update agent instance and configuration yaml file."""
    agent_updates = {}
    if agent_dto.description is not None and agent_dto.description != "":
        agent_updates['info.description'] = agent_dto.description
        agent.agent_model.info['description'] = agent_dto.description
    if agent_dto.prompt is not None:
        prompt_dto = agent_dto.prompt
        if prompt_dto.target is not None:
            agent_updates['profile.target'] = agent_dto.prompt.target
            agent.agent_model.profile['target'] = agent_dto.prompt.target
        if prompt_dto.introduction is not None:
            agent_updates['profile.introduction'] = agent_dto.prompt.introduction
            agent.agent_model.profile['introduction'] = agent_dto.prompt.introduction
        if prompt_dto.instruction is not None:
            agent_updates['profile.instruction'] = agent_dto.prompt.instruction
            agent.agent_model.profile['instruction'] = agent_dto.prompt.instruction
    if agent_dto.llm is not None:
        llm_dto = agent_dto.llm
        if hasattr(agent, 'llm_name'):
            agent.llm_name = agent_dto.llm.id
        if llm_dto.id:
            agent_updates['profile.llm_model.name'] = agent_dto.llm.id
            agent.agent_model.profile.get('llm_model')['name'] = agent_dto.llm.id
        if llm_dto.temperature:
            agent_updates['profile.llm_model.temperature'] = agent_dto.llm.temperature
            agent.agent_model.profile.get('llm_model')['temperature'] = agent_dto.llm.temperature
        if llm_dto.model_name:
            agent_updates['profile.llm_model.model_name'] = agent_dto.llm.model_name[0]
            agent.agent_model.profile.get('llm_model')['model_name'] = agent_dto.llm.model_name[0]
    if agent_dto.tool is not None:
        tool_dto_list: List[ToolDTO] = agent_dto.tool
        tool_name_list = [tool_dto.id for tool_dto in tool_dto_list]
        if hasattr(agent, 'tool_names'):
            agent.tool_names = tool_name_list
        agent_updates['action.tool'] = tool_name_list
        agent.agent_model.action['tool'] = tool_name_list
    if agent_dto.knowledge is not None:
        knowledge_dto_list: List[KnowledgeDTO] = agent_dto.knowledge
        knowledge_name_list = [knowledge_dto.id for knowledge_dto in knowledge_dto_list]
        if hasattr(agent, 'knowledge_names'):
            agent.knowledge_names = knowledge_name_list
        agent_updates['action.knowledge'] = knowledge_name_list
        agent.agent_model.action['knowledge'] = knowledge_name_list
    if agent_updates:
        update_nested_yaml_value(agent_config_path, agent_updates)


def update_agent_product_config(agent_product: Product, agent_dto: AgentDTO, product_config_path: str) -> None:
    """Update agent product instance and configuration yaml file."""
    if agent_dto is None or agent_dto.opening_speech is None:
        return
    agent_product.opening_speech = agent_dto.opening_speech
    update_nested_yaml_value(product_config_path, {'opening_speech': agent_dto.opening_speech})


def assemble_tool_input(agent: Agent, agent_input: str) -> dict:
    """Assemble tool input parameters for the agent invocation.

    Args:
        agent (Agent): The agent instance.
        agent_input (str): The agent input string.

    Returns:
        dict: The tool input parameters dictionary.
    """
    tool_id_list = agent.agent_model.action.get('tool', [])
    tool_input_dict = {}
    for tool_id in tool_id_list:
        tool: Tool = ToolManager().get_instance_obj(tool_id)
        if tool is None:
            continue
        if len(tool.input_keys) > 0:
            input_key = tool.input_keys[0]
            tool_input_dict[input_key] = agent_input
    return tool_input_dict


def validate_and_assemble_agent_input(agent_id: str, session_id: str, input: str, chat_history: list = None) -> dict:
    """Validate and assemble agent input.

    Args:
        agent_id (str): The agent id.
        session_id (str): The session id.
        input (str): The agent input string.
        chat_history (list): The chat history list.

    Returns:
        dict: The agent input dictionary.
    """
    if agent_id is None or session_id is None:
        raise ValueError("Agent id or session id cannot be None.")
    agent: Agent = AgentManager().get_instance_obj(agent_id)
    if agent is None:
        raise ValueError("The agent instance corresponding to the agent id cannot be found.")
    tool_input_dict = assemble_tool_input(agent, input)

    agent_input_dict = {
        'input': input,
        'chat_history': chat_history,
        'agent_id': agent_id,
        'session_id': session_id
    }

    agent_input_dict.update(tool_input_dict)

    return agent_input_dict



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 21:11
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_service.py
from datetime import datetime
import json
import time
import os
from typing import List, Tuple, Iterator, AsyncIterator

from agentuniverse.agent.agent import Agent
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.agent_serve.web.request_task import RequestTask
from agentuniverse.agent_serve.web.web_util import agent_run_queue, async_agent_run_queue
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.util.monitor.monitor import Monitor
from agentuniverse_product.base.product_manager import Product
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.base.util.yaml_util import write_yaml_file
from agentuniverse_product.service.message_service.message_service import MessageService
from agentuniverse_product.service.model.agent_dto import AgentDTO
from agentuniverse_product.service.model.message_dto import MessageDTO
from agentuniverse_product.service.model.session_dto import SessionDTO
from agentuniverse_product.service.session_service.session_service import SessionService
from agentuniverse_product.service.util.agent_util import validate_create_agent_parameters, \
    assemble_product_config_data, assemble_agent_config_data, assemble_agent_dto, update_agent_product_config, \
    update_agent_config, register_agent, register_product, validate_and_assemble_agent_input
from agentuniverse_product.service.util.common_util import get_core_path


class AgentService:
    """Agent Service for aU-product."""

    @staticmethod
    def get_agent_list() -> List[AgentDTO]:
        """Get all agents.

        Returns:
            List[AgentDTO]: List of AgentDTOs.
        """
        res = []
        product_list: List[Product] = ProductManager().get_instance_obj_list()
        if len(product_list) < 1:
            return res
        for product in product_list:
            if product.type == ComponentEnum.AGENT.value:
                agent_dto = AgentDTO(nickname=product.nickname, avatar=product.avatar, id=product.id,
                                     opening_speech=product.opening_speech, mtime=product.get_mtime())
                agent = product.instance
                agent_model: AgentModel = agent.agent_model
                agent_dto.description = agent_model.info.get('description', '')
                res.append(agent_dto)
        return res

    @staticmethod
    def get_agent_detail(id: str) -> AgentDTO | None:
        """Get agent detail by agent id.

        Returns:
            AgentDTO | None: AgentDTO or None.
        """
        product: Product = ProductManager().get_instance_obj(id)
        agent: Agent = AgentManager().get_instance_obj(id)
        if agent is None:
            raise ValueError("The agent instance corresponding to the agent id cannot be found.")
        agent_dto = AgentDTO(id=id,
                             nickname=product.nickname if product else '',
                             avatar=product.avatar if product else '',
                             opening_speech=product.opening_speech if product else '',
                             mtime=product.get_mtime())
        # assemble agent dto
        return assemble_agent_dto(agent, agent_dto)

    @staticmethod
    def update_agent(agent_dto: AgentDTO) -> None:
        """Update agent detail."""
        if agent_dto.id is None:
            raise ValueError("Agent id cannot be None.")
        product: Product = ProductManager().get_instance_obj(agent_dto.id)
        agent: Agent = AgentManager().get_instance_obj(agent_dto.id)
        if agent is None:
            raise ValueError("The agent instance corresponding to the agent id cannot be found.")
        # update agent product yaml configuration file
        if product:
            update_agent_product_config(product, agent_dto, product.component_config_path)
        # update agent yaml configuration file
        update_agent_config(agent, agent_dto, agent.component_config_path)

    @staticmethod
    def chat(agent_id: str, session_id: str, input: str) -> dict:
        """Chat with agent and get response.

        Args:
            agent_id (str): Agent id.
            session_id (str): Session id.
            input (str): Query string.
        Returns:
            dict: Response.
        """
        if agent_id is None or session_id is None:
            raise ValueError("Agent id or session id cannot be None.")
        agent: Agent = AgentManager().get_instance_obj(agent_id)
        if agent is None:
            raise ValueError("The agent instance corresponding to the agent id cannot be found.")

        # init the invocation chain and token usage of the monitor module
        Monitor.init_invocation_chain()
        Monitor.init_token_usage()

        # invoke agent
        start_time = time.time()
        output_object: OutputObject = agent.run(input=input, session_id=session_id)
        end_time = time.time()
        # calculate response time
        response_time = round((end_time - start_time) * 1000, 2)
        output = output_object.get_data('output')

        # get and clear invocation chain and token usage.
        invocation_chain = Monitor.get_invocation_chain()
        token_usage = Monitor.get_token_usage()
        Monitor.clear_invocation_chain()
        Monitor.clear_token_usage()

        # add agent chat history
        session_id, message_id = AgentService().add_agent_chat_history(agent_id, session_id, input, output,
                                                                       datetime.fromtimestamp(end_time))

        return {'response_time': response_time, 'message_id': message_id, 'session_id': session_id, 'output': output,
                'start_time': datetime.fromtimestamp(start_time).strftime("%Y-%m-%d %H:%M:%S"),
                'end_time': datetime.fromtimestamp(end_time).strftime("%Y-%m-%d %H:%M:%S"),
                'invocation_chain': invocation_chain, 'token_usage': token_usage}

    @staticmethod
    def stream_chat(agent_id: str, session_id: str, input: str) -> Iterator:
        """Stream chat with agent and get response.

        Args:
            agent_id (str): Agent id.
            session_id (str): Session id.
            input (str): Query string.
        Returns:
            Iterator: Response.
        """
        agent_input_dict = validate_and_assemble_agent_input(agent_id, session_id, input)

        # init the invocation chain and token usage of the monitor module
        Monitor.init_invocation_chain()
        Monitor.init_token_usage()

        # invoke agent
        start_time = time.time()
        task = RequestTask(agent_run_queue, False,
                           **agent_input_dict)

        # generate iterator
        for chunk in task.stream_run():
            chunk_dict = json.loads(chunk.replace("data:", "", 1))
            if "process" in chunk_dict:
                data = chunk_dict['process'].get('data')
                if data:
                    yield_type = 'token' if 'chunk' in data else 'intermediate_steps'
                    yield {'output': data.get('chunk' if yield_type == 'token' else 'output'),
                           'type': yield_type,
                           'agent_id': data.get('agent_info', {}).get('name', '')}
            elif "result" in chunk_dict or "error" in chunk_dict:
                end_time = time.time()
                # calculate response time
                response_time = round((end_time - start_time) * 1000, 2)
                if "result" in chunk_dict:
                    output = chunk_dict['result'].get('output')
                    # add agent chat history
                    session_id, message_id = AgentService().add_agent_chat_history(
                        agent_id, session_id, input, output, datetime.fromtimestamp(end_time)
                    )
                    # get and clear contexts
                    invocation_chain = Monitor.get_invocation_chain()
                    token_usage = Monitor.get_token_usage()
                    Monitor.clear_invocation_chain()
                    Monitor.clear_token_usage()
                    # return final yield
                    yield {
                        'response_time': response_time,
                        'message_id': message_id,
                        'session_id': session_id,
                        'output': output,
                        'start_time': datetime.fromtimestamp(start_time).strftime("%Y-%m-%d %H:%M:%S"),
                        'end_time': datetime.fromtimestamp(end_time).strftime("%Y-%m-%d %H:%M:%S"),
                        'invocation_chain': invocation_chain,
                        'token_usage': token_usage,
                        'type': 'final_result'
                    }
                else:
                    # clear contexts
                    Monitor.clear_invocation_chain()
                    Monitor.clear_token_usage()
                    yield {'error': chunk_dict['error'], 'type': 'error'}

    @staticmethod
    async def async_stream_chat(agent_id: str, session_id: str, input: str) -> AsyncIterator:
        """Asynchronously stream chat with agent and get response.

        Args:
            agent_id (str): Agent id.
            session_id (str): Session id.
            input (str): Query string.
        Returns:
            AsyncIterator: Response.
        """
        agent_input_dict = validate_and_assemble_agent_input(agent_id, session_id, input)

        # init the invocation chain and token usage of the monitor module
        Monitor.init_invocation_chain()
        Monitor.init_token_usage()

        # invoke agent
        start_time = time.time()
        task = RequestTask(async_agent_run_queue, False, **agent_input_dict)

        # generate async iterator
        async for chunk in task.async_stream_run():
            chunk_dict = json.loads(chunk.replace("data:", "", 1))
            if "process" in chunk_dict:
                data = chunk_dict['process'].get('data')
                if data:
                    yield_type = 'token' if 'chunk' in data else 'intermediate_steps'
                    yield {'output': data.get('chunk' if yield_type == 'token' else 'output'),
                           'type': yield_type,
                           'agent_id': data.get('agent_info', {}).get('name', '')}
            elif "result" in chunk_dict or "error" in chunk_dict:
                end_time = time.time()
                # calculate response time
                response_time = round((end_time - start_time) * 1000, 2)
                if "result" in chunk_dict:
                    output = chunk_dict['result'].get('output')
                    # add agent chat history
                    session_id, message_id = AgentService().add_agent_chat_history(
                        agent_id, session_id, input, output, datetime.fromtimestamp(end_time)
                    )
                    # get and clear contexts
                    invocation_chain = Monitor.get_invocation_chain()
                    token_usage = Monitor.get_token_usage()
                    Monitor.clear_invocation_chain()
                    Monitor.clear_token_usage()
                    # return final yield
                    yield {
                        'response_time': response_time,
                        'message_id': message_id,
                        'session_id': session_id,
                        'output': output,
                        'start_time': datetime.fromtimestamp(start_time).strftime("%Y-%m-%d %H:%M:%S"),
                        'end_time': datetime.fromtimestamp(end_time).strftime("%Y-%m-%d %H:%M:%S"),
                        'invocation_chain': invocation_chain,
                        'token_usage': token_usage,
                        'type': 'final_result'
                    }
                else:
                    # clear contexts
                    Monitor.clear_invocation_chain()
                    Monitor.clear_token_usage()
                    yield {'error': chunk_dict['error'], 'type': 'error'}

    @staticmethod
    def create_agent(agent_dto: AgentDTO) -> str:
        """Create an agent instance.

        Args:
            agent_dto (AgentDTO): AgentDTO.
        Returns:
            str: Agent id.
        """
        # validate parameters
        validate_create_agent_parameters(agent_dto)

        # assemble product config data
        product_config_data = assemble_product_config_data(agent_dto)

        # write product YAML file
        product_file_name = f"{agent_dto.id}_product"

        path = get_core_path()
        product_file_path = path / 'product' / 'agent' / f"{product_file_name}.yaml" if path \
            else os.path.join('..', '..', 'platform', 'difizen', 'product', 'agent', f"{product_file_name}.yaml")
        write_yaml_file(str(product_file_path), product_config_data)

        # assemble agent config data
        agent_config_data = assemble_agent_config_data(agent_dto)

        # write agent YAML file
        agent_file_path = path / 'agent' / f"{agent_dto.id}.yaml" if path \
            else os.path.join('..', '..', 'intelligence', 'agentic', 'agent', f"{agent_dto.id}.yaml")
        write_yaml_file(str(agent_file_path), agent_config_data)

        # register product and agent instance
        register_agent(str(agent_file_path))
        register_product(str(product_file_path))
        return agent_dto.id

    @staticmethod
    def get_agent_chat_history(session_id: str) -> list:
        """Get agent chat history from session db table by session id."""
        # newest top5
        session_dto: SessionDTO = SessionService().get_session_detail(session_id, 5)
        chat_history = []
        if session_dto:
            messages: List[MessageDTO] = session_dto.messages
            if len(messages) > 0:
                for message in messages:
                    content: str = message.content
                    if content is not None:
                        chat_history.extend(json.loads(content))
        return chat_history

    @staticmethod
    def add_agent_chat_history(agent_id: str, session_id: str,
                               input: str, output: str, cur_time: datetime) -> Tuple[str, int]:
        """Add agent chat history to session and message db tables."""
        content = json.dumps([{'type': 'human', 'content': input}, {'type': 'ai', 'content': output}],
                             ensure_ascii=False)
        session_id = SessionService.update_session(session_id, agent_id, cur_time)
        message_id = MessageService.add_message(session_id, content, cur_time)
        return session_id, message_id



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/28 16:54
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: hierarchical_text_spliter.py
import re
import uuid
from typing import List, Optional

from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import \
    DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class HierarchicalRegexTextSplitter(DocProcessor):
    merge_first: bool = False
    hierarchical_index: List[dict] = [
        {
            "reg_exp": "第[零一二三四五六七八九十百千]+章",
            "need_summary": True
        },
        {
            "reg_exp": "第[零一二三四五六七八九十百千]+节",
            "need_summary": False
        }
    ]
    summary_agent: str = "simple_summary_agent"
    llm: Optional[dict] = None

    def _hierarchical_split_single_doc(self, doc: Document) -> List[Document]:
        hierarchy = {}
        levels = [(f"item_{i}", re.compile(index['reg_exp'])) for i, index in enumerate(self.hierarchical_index)]
        current_level = {level[0]: None for level in levels}
        current_level['root'] = None
        root = Document(
            text='',
            metadata={
                'hierarchical_parent': 'root'
            }
        )
        hierarchy['root'] = root

        last_inserted_id = 'root'

        # build hierarchical doc tree
        for line in doc.text.splitlines():
            line = line.strip()
            if not line:
                continue
            for level_name, pattern in levels:
                match = pattern.match(line)
                if match:
                    _ptr = levels.index((level_name, pattern))
                    parent_level = \
                        levels[_ptr - 1][0] if _ptr > 0 else 'root'
                    parent_id = current_level[parent_level].id if \
                        current_level[parent_level] else 'root'
                    # use uuid4 to generate a random id, because chapter name can always duplicate
                    node = Document(
                        id=str(uuid.uuid4()),
                        text=line,
                        metadata={
                            'hierarchical_parent': parent_id,
                            'hierarchical_info': match.group(0) if match else '',
                            'hierarchical_level': _ptr
                        }
                    )
                    hierarchy[node.id] = node
                    current_level[level_name] = node
                    last_inserted_id = node.id
                    break
            else:
                hierarchy[last_inserted_id].text += "\n" + line
            # Append text to the last inserted node's text and all its parent
            _inserted_id = hierarchy[last_inserted_id].metadata['hierarchical_parent']
            while _inserted_id != 'root':
                hierarchy[_inserted_id].text += "\n" + line
                _inserted_id = hierarchy[_inserted_id].metadata['hierarchical_parent']
        hierarchy.pop("root")
        hierarchical_docs = []

        # summary the text if docs hierarchical_level need summary
        for k, v in hierarchy.items():
            hierarchical_level = v.metadata['hierarchical_level']
            if self.hierarchical_index[hierarchical_level]['need_summary']:
                agent = AgentManager().get_instance_obj(self.summary_agent)
                if self.llm:
                    agent.agent_model.profile['llm_model'] = self.llm
                else:
                    agent.agent_model.profile['llm_model'] = {
                        "name": "__default_instance__"}
                agent_result = agent.run(input=v.text)
                v.text = agent_result.output
            hierarchical_docs.append(v)

        return hierarchical_docs

    def _process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        # merge all documents first
        merged_docs = origin_docs
        if self.merge_first and len(origin_docs) > 0:
            for _doc in origin_docs[1:]:
                origin_docs[0].text += "\n" + _doc.text
            merged_docs = [origin_docs[0]]
        hierarchical_docs = []
        for _doc in merged_docs:
            hierarchical_docs.extend(self._hierarchical_split_single_doc(_doc))
        return hierarchical_docs

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(doc_processor_configer)
        if hasattr(doc_processor_configer, "hierarchical_index"):
            self.hierarchical_index = doc_processor_configer.hierarchical_index
        if hasattr(doc_processor_configer, "merge_first"):
            self.merge_first = doc_processor_configer.merge_first
        if hasattr(doc_processor_configer, "llm"):
            self.llm = doc_processor_configer.llm
        if hasattr(doc_processor_configer, "summary_agent"):
            self.summary_agent = doc_processor_configer.summary_agent
        return self




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 13:52
# @Author  : weizjajj
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: qwen_openai_style_llm.py
from typing import Optional, Any, Union, Iterator, AsyncIterator

from dashscope import get_tokenizer
from pydantic import Field

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.openai_style_llm import OpenAIStyleLLM

QWen_Max_CONTEXT_LENGTH = {
    "qwen-turbo": 6000,
    "qwen-plus": 300000,
    "qwen-max": 6000,
    "qwen-max-0428": 6000,
    "qwen-max-0403": 6000,
    "qwen-max-0107": 6000,
    "qwen-max-longcontext": 28000
}


class QWenOpenAIStyleLLM(OpenAIStyleLLM):
    """
        QWen OpenAI style LLM
        Args:
            api_key: API key for the model ,from dashscope : DASHSCOPE_API_KEY
            api_base: API base URL for the model, from dashscope : DASHSCOPE_API_BASE
    """

    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("DASHSCOPE_API_KEY"))
    api_base: Optional[str] = "https://dashscope.aliyuncs.com/compatible-mode/v1"
    proxy: Optional[str] = Field(default_factory=lambda: get_from_env("DASHSCOPE_PROXY"))
    organization: Optional[str] = Field(default_factory=lambda: get_from_env("DASHSCOPE_ORGANIZATION"))

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """ The call method of the LLM.

        Users can customize how the model interacts by overriding call method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return super()._call(messages, **kwargs)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """ The async call method of the LLM.

        Users can customize how the model interacts by overriding acall method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return await super()._acall(messages, **kwargs)

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return QWen_Max_CONTEXT_LENGTH.get(self.model_name, 8000)

    def get_num_tokens(self, text: str) -> int:
        tokenizer = get_tokenizer(self.model_name)
        return len(tokenizer.encode(text))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/27 15:34
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: plugin_product.py
from typing import Optional, List

from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_configer import ProductConfiger


class PluginProduct(Product):
    """The basic class of the plugin product."""

    toolset: Optional[List[str]] = list()
    openapi_desc: Optional[str] = None

    def initialize_by_component_configer(self,
                                         product_configer: ProductConfiger) -> 'PluginProduct':
        """Initialize the Product by the ProductConfiger object.

        Args:
            product_configer(ProductConfiger): A configer contains the product basic info.
        Returns:
            PluginProduct: A PluginProduct instance.
        """
        if product_configer.nickname:
            self.nickname = product_configer.nickname
        if product_configer.id:
            self.id = product_configer.id
        if product_configer.type:
            self.type = product_configer.type
        if product_configer.avatar:
            self.avatar = product_configer.avatar
        if product_configer.description:
            self.description = product_configer.description
        if hasattr(product_configer, 'toolset') and product_configer.toolset:
            self.toolset = product_configer.toolset
        if hasattr(product_configer, 'openapi_desc') and product_configer.openapi_desc:
            self.openapi_desc = product_configer.openapi_desc
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 18:23
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: default_memory.py
from agentuniverse.agent.memory.chat_memory import ChatMemory
from agentuniverse.llm.default.default_openai_llm import DefaultOpenAILLM


class DefaultMemory(ChatMemory):
    """The aU default memory module."""

    def __init__(self, **kwargs):
        """The __init__ method.

        Some parameters, such as name/description/type/memory_key,
        are injected into this class by the default_memory.yaml configuration.


        Args:
            llm (LLM): the LLM instance used by this memory.
            default memory uses OpenAILLM(gpt-3.5-turbo) object as the memory llm.
        """
        super().__init__(**kwargs)
        self.llm = DefaultOpenAILLM(model_name="gpt-4o")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 14:04
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: sls_sink.py

import queue
import threading
import time
from typing import List, Optional

from aliyun.log.logclient import LogClient
from aliyun.log.logitem import LogItem
from aliyun.log.putlogsrequest import PutLogsRequest
from aliyun.log.putlogsresponse import PutLogsResponse
import loguru


class SlsSender:
    """A class to send log to aliyun simple log server."""

    def __init__(
            self,
            project: str,
            log_store: str,
            endpoint: str,
            access_key_id: str,
            access_key_secret: str,
            queue_max_size: int,
            send_interval: float):
        """Initialize a sls sender.

        Args:
            project (`str`):
                Project name of aliyun sls.
            log_store (`str`):
                Log store of aliyun sls.
            endpoint (`str`):
                Endpoint of aliyun sls.
            access_key_id (`str`):
                Project name of aliyun sls.
            access_key_secret (`str`):
                Project name of aliyun sls.
            queue_max_size (`int`):
                Log queue max size, sls sender use a queue to save the logs
                to be sent, a separate thread will upload logs to aliyun sls
                periodically.
            send_interval (`float`):
                Interval of the separate thread sending logs to aliyun sls.
        """
        self.project = project
        self.log_store = log_store
        self.client = LogClient(endpoint, access_key_id, access_key_secret)
        self.log_queue = queue.Queue(queue_max_size)
        self.send_interval = send_interval

        self.send_thread_stop_event = threading.Event()
        self.send_thread = None
        self._logger = loguru.logger

    def _send_put_logs_request(self,
                               log_item_list: List[LogItem],
                               topic: str = "",
                               source: str = "") -> Optional[PutLogsResponse]:
        """Send a batch of logs to aliyun sls.

        Args:
            log_item_list (`List[LogItem]`):
                A list of log items to be sent to aliyun sls.
            topic (`str`, defaults to `""`):
                An attribute used to identify a group of logs in aliyun sls.
            source (`str`, defaults to `""`):
                An identifier that allows to discern the source of the log.

        Returns:
            If logs sent successfully, returns a PutLogsResponse, else returns
            none and log the error in local log file.
        """
        try:
            put_request = PutLogsRequest(self.project, self.log_store, topic,
                                         source, log_item_list)
            put_response = self.client.put_logs(put_request)
        except Exception as e:
            self._logger.error(
                f"send single log to sls failed: {str(e)}", )
            return None
        return put_response

    def send_single_log(self,
                        message: str,
                        topic: str = "",
                        source: str = "") -> Optional[PutLogsResponse]:
        """Send a single log to aliyun sls.

        Args:
            message (`str`):
                The message to be sent to aliyun sls.
            topic (`str`, defaults to `""`):
                An attribute used to identify a group of logs in aliyun sls.
            source (`str`, defaults to `""`):
                An identifier that allows to discern the source of the log.

        Returns:
            If logs sent successfully, returns a PutLogsResponse, else returns
            none and log the error in local log file.
        """
        log_item_list = list()
        log_item = LogItem()
        log_item.set_contents(message)
        log_item_list.append(log_item)
        return self._send_put_logs_request(log_item_list, topic, source)

    def put_log_queue(self, log_item: LogItem):
        """Put a single log item into the waiting queue.

        Args:
            log_item (`LogItem`):
                The log item to be put into the queue.
        """
        try:
            self.log_queue.put(log_item, block=False)
        except queue.Full:
            self._logger.error("sls log queue is full, discard new log")

    def batch_send(self,
                   topic: str = "",
                   source: str = "") -> Optional[PutLogsResponse]:
        """Send all log items in waiting queue to aliyun sls.

        Args:
            topic (`str`, defaults to `""`):
                An attribute used to identify a group of logs in aliyun sls.
            source (`str`, defaults to `""`):
                An identifier that allows to discern the source of the log.

        Returns:
            If logs sent successfully, returns a PutLogsResponse, else returns
            none and log the error in local log file.
        """

        # Get all log items in waiting queue.
        size = self.log_queue.qsize()
        log_item_list = []
        if self.log_queue is not None and size > 0:
            for i in range(size):
                try:
                    log_item = self.log_queue.get(block=False)
                except queue.Empty:
                    self._logger.error(
                        "sls log queue shorter than expected, "
                        "all logs have been sent")
                    break
                log_item_list.append(log_item)

        # Send all log items to aliyun sls.
        length = len(log_item_list)
        if length > 0:
            return self._send_put_logs_request(log_item_list, topic, source)
        return None

    def start_batch_send_thread(self):
        """Start the log sending thread."""
        if self.send_thread is None or not self.send_thread.is_alive():
            self.send_thread_stop_event.clear()
            self.send_thread = threading.Thread(
                target=self._schedule_send_log,
                name="loop_send_log_thread", daemon=True)
            self.send_thread.start()

    def stop_batch_send_thread(self):
        """Stop the log sending thread."""
        if self.send_thread is not None:
            self.send_thread_stop_event.set()
            self.send_thread.join()
            self.send_thread = None

    def _schedule_send_log(self):
        """Create an infinite loop uploading logs in queue to aliyun sls."""
        while not self.send_thread_stop_event.is_set():
            self.batch_send()
            time.sleep(self.send_interval)


class SlsSink:
    """A custom loguru sink used to send logs to aliyun sls."""

    def __init__(self, sls_sender: SlsSender):
        """Initialize the sls sink."""
        self.sls_sender = sls_sender

    def __call__(self, message):
        """Construct the message to a sls log item and put it to the queue
         waiting for sls sender to send to the aliyun sls."""
        log_item = LogItem()
        log_item.set_time(int(message.record["time"].timestamp()))
        log_content = list()
        log_content.append(("content", str(message)))
        log_item.set_contents(log_content)
        self.sls_sender.put_log_queue(log_item)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 19:39
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: graph.py
from typing import Optional, Any

import networkx as nx

from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import Node
from agentuniverse.workflow.node.node_constant import NODE_CLS_MAPPING
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.workflow_output import WorkflowOutput


class Graph(nx.DiGraph):
    """The basic class of the graph."""

    def build(self, workflow_id: str, config: dict) -> 'Graph':
        """Build the graph."""
        nodes_config = config.get('nodes')
        if not nodes_config:
            raise ValueError('The nodes configuration is empty')
        for node_config in nodes_config:
            self._add_graph_node(workflow_id, node_config)
        edges_config = config.get('edges')
        if not edges_config:
            raise ValueError('The edges configuration is empty')
        for edge_config in edges_config:
            self._add_graph_edge(edge_config)
        if not nx.is_directed_acyclic_graph(self):
            raise ValueError("The provided configuration does not form a DAG.")
        return self

    def _add_graph_node(self, workflow_id: str, node_config: dict) -> None:
        """Add a node to the graph.

        Args:
            workflow_id: The workflow id.
            node_config: The node configuration.
        """
        if node_config.get('type') not in NodeEnum.to_value_list():
            raise ValueError('The node type is not supported')
        node_id = node_config.get('id')
        if self.has_node(node_id):
            return
        node_cls = NODE_CLS_MAPPING[node_config.get('type')]
        node_type = node_config.pop('type')
        node_instance = node_cls(type=NodeEnum.from_value(node_type), workflow_id=workflow_id,
                                 **node_config)
        node_config['type'] = node_type
        self.add_node(node_id, instance=node_instance, type=node_type)

    def _add_graph_edge(self, edge_config: dict) -> None:
        """Add a edge to the graph.

        Args:
            edge_config: The edge configuration.
        """
        self.add_edge(edge_config.get('source_node_id'), edge_config.get('target_node_id'),
                      source_handler=edge_config.get('source_handler'))

    def run(self, workflow_output: WorkflowOutput) -> None:
        """Run the graph.

        Args:
            workflow_output: The workflow output.
        """
        sorted_nodes = list(nx.topological_sort(self))
        predecessor_node: Node | None = None
        while True:
            next_node = self._get_next_node(workflow_output, sorted_nodes, predecessor_node)
            if not next_node:
                break
            if self._has_node_been_executed(workflow_output, next_node.id):
                predecessor_node = next_node
                continue
            self._run_node(cur_node=next_node, workflow_output=workflow_output)
            if next_node.type == NodeEnum.END:
                break
            predecessor_node = next_node

    def _get_next_node(self, workflow_output: WorkflowOutput, nodes: Any,
                       predecessor_node: Optional[Node] = None) -> Optional[Node]:
        """Get the next node in the graph.

        Args:
            workflow_output: The workflow output.
            nodes: The nodes in the graph.
            predecessor_node: The predecessor node.
        Returns:
            The next node in the graph.
        """
        if not predecessor_node:
            for node_id in nodes:
                if self.nodes[node_id]['type'] == NodeEnum.START.value:
                    return self.nodes[node_id]['instance']
        else:
            predecessor_node_output: NodeOutput = workflow_output.workflow_node_results.get(
                predecessor_node.id,
                None
            )
            source_handler = predecessor_node_output.edge_source_handler if predecessor_node_output else None
            successors = self.successors(predecessor_node.id)
            if not successors:
                return None
            successors = list(successors)
            if source_handler:
                for s in successors:
                    if self.get_edge_data(predecessor_node.id, s).get('source_handler') == source_handler:
                        return self.nodes[s]['instance']
            else:
                return self.nodes[successors[0]]['instance']

    @staticmethod
    def _has_node_been_executed(workflow_output: WorkflowOutput, node_id: str) -> bool:
        """Check if the node has been executed.

        Args:
            workflow_output: The workflow output.
            node_id: The node id.
        Returns:
            True if the node has been executed, False otherwise.
        """
        return node_id in workflow_output.workflow_node_results

    @staticmethod
    def _run_node(cur_node: Node = None,
                  workflow_output: WorkflowOutput = None) -> None:
        """Run the node in the graph.

        Args:
            cur_node: The current node in the graph.
            workflow_output: The workflow output.
        """
        node_output = cur_node.run(workflow_output)
        workflow_output.workflow_node_results[cur_node.id] = node_output



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/17 20:41
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: executing_agent.py
from agentuniverse.agent.template.executing_agent_template import ExecutingAgentTemplate


class ExecutingAgent(ExecutingAgentTemplate):
    """Executing Agent class."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/21 15:01
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: start_node.py
from typing import List

from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node_config import NodeOutputParams
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.node.node import Node
from agentuniverse.workflow.workflow_output import WorkflowOutput


class StartNode(Node):
    """The basic class of the start node."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.START

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        start_params: dict = workflow_output.workflow_start_params
        start_val = start_params.get('input', '')
        output_params: List[NodeOutputParams] = self._data.outputs
        output_params[0].value = start_val
        workflow_output.workflow_parameters[self.id] = output_params
        return NodeOutput(node_id=self.id, status=NodeStatusEnum.SUCCEEDED, result=output_params)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/26 18:11
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: markdown_reader.py
from typing import Union
from pathlib import Path
from typing import List, Optional, Dict

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document




class MarkdownReader(Reader):
    """Docx reader."""

    def _load_data(self, file: Union[str, Path], ext_info: Optional[Dict] = None) -> List[Document]:
        """Parse the markdown file.

        Note:
            The markdown file cannot be process in pagination.
            `unstructured` and `markdown` is required to read markdown files:
            `pip install unstructured` and `pip install markdown`
        """
        try:
            from langchain_community.document_loaders import \
                UnstructuredMarkdownLoader
        except ImportError:
            raise ImportError(
                "markdown and unstructured is required to read markdown files: "
                "please run `pip install unstructured` and `pip install markdown`"
            )

        if isinstance(file, str):
            file = Path(file)
        data = UnstructuredMarkdownLoader(file).load()
        metadata = {"file_name": file.name}
        if ext_info is not None:
            metadata.update(ext_info)

        return [Document(text=data[0].page_content, metadata=metadata or {})]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/19 19:19
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: document.py
import uuid
from typing import Dict, Any, Optional, List, Set

from langchain_core.documents.base import Document as LCDocument
from pydantic import BaseModel, Field, model_validator


class Document(BaseModel):
    """The basic class for the document.

    Attributes:
        id (str): Unique identifier for the document.
        text (Optional[str]): The content of the document.
        metadata (Dict[str, Any]): Metadata associated with the document.
        embedding (List[float]): Embedding data associated with the document
    """
    class Config:
        arbitrary_types_allowed = True

    id: str = None
    text: Optional[str] = ""
    metadata: Optional[Dict[str, Any]] = None
    embedding: List[float] = Field(default_factory=list)
    keywords: Set[str] = Field(default_factory=set)

    @model_validator(mode='before')
    def create_id(cls, values):
        text: str = values.get('text', '')
        if not values.get('id'):
            values['id'] = str(uuid.uuid5(uuid.NAMESPACE_URL, text))
        return values

    def as_langchain(self) -> LCDocument:
        """Convert to LangChain document format."""
        metadata = self.metadata or {}
        return LCDocument(page_content=self.text, metadata=metadata)

    @staticmethod
    def as_langchain_list(document_list) -> List[LCDocument]:
        """Convert agentUniverse(aU) document list to langchain document list """
        langchain_document_list = []
        if document_list is None:
            return langchain_document_list
        for document in document_list:
            langchain_document_list.append(LCDocument(page_content=document.text, metadata=document.metadata))
        return langchain_document_list

    @staticmethod
    def from_langchain_list(lc_document_list: List[LCDocument]):
        """Convert langchain document list to agentUniverse(aU) document list """
        document_list = []
        if lc_document_list is None:
            return document_list
        for lc_document in lc_document_list:
            document_list.append(Document(text=lc_document.page_content, metadata=lc_document.metadata))
        return document_list



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 14:30
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: reader.py
from abc import abstractmethod
from typing import List, Any, Optional

from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class Reader(ComponentBase):
    """The basic class for the knowledge reader."""
    component_type: ComponentEnum = ComponentEnum.READER
    name: Optional[str] = None
    description: Optional[str] = None

    def load_data(self, *args: Any, **kwargs: Any) -> List[Document]:
        """Load data from the input params."""
        return self._load_data(*args, **kwargs)

    @abstractmethod
    def _load_data(self, *args: Any, **kwargs: Any) -> List[Document]:
        """Load data from the input params."""

    def _initialize_by_component_configer(self,
                                         reader_configer: ComponentConfiger) \
            -> 'Reader':
        """Initialize the reader by the ComponentConfiger object.

        Args:
            reader_configer(ComponentConfiger): A configer contains reader
            basic info.
        Returns:
            Reader: A reader instance.
        """
        self.name = reader_configer.name
        self.description = reader_configer.description
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/15 11:42
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: enum.py
import enum
from enum import Enum


@enum.unique
class MemoryTypeEnum(Enum):
    SHORT_TERM = 'short_term'
    LONG_TERM = 'long_term'


@enum.unique
class ChatMessageEnum(Enum):
    SYSTEM = 'system'
    HUMAN = 'human'
    AI = 'ai'



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 16:17
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: app_configer.py
from typing import Optional
from agentuniverse.base.config.configer import Configer


class AppConfiger(object):
    """The AppConfiger class, which is used to load and manage the application configuration."""

    def __init__(self):
        """Initialize the AppConfiger."""
        self.__configer: Optional[Configer] = None
        self.__base_info_appname: Optional[str] = None
        self.__core_default_package_list: Optional[list[str]] = None
        self.__core_agent_package_list: Optional[list[str]] = None
        self.__core_knowledge_package_list: Optional[list[str]] = None
        self.__core_llm_package_list: Optional[list[str]] = None
        self.__core_planner_package_list: Optional[list[str]] = None
        self.__core_tool_package_list: Optional[list[str]] = None
        self.__core_memory_package_list: Optional[list[str]] = None
        self.__core_service_package_list: Optional[list[str]] = None
        self.__core_sqldb_wrapper_package_list: Optional[list[str]] = None
        self.__core_prompt_package_list: Optional[list[str]] = None
        self.__core_product_package_list: Optional[list[str]] = None
        self.__core_workflow_package_list: Optional[list[str]] = None
        self.__core_embedding_package_list: Optional[list[str]] = None
        self.__core_doc_processor_package_list: Optional[list[str]] = None
        self.__core_reader_package_list: Optional[list[str]] = None
        self.__core_store_package_list: Optional[list[str]] = None
        self.__core_rag_router_package_list: Optional[list[str]] = None
        self.__core_query_paraphraser_package_list: Optional[list[str]] = None
        self.__core_memory_compressor_package_list: Optional[list[str]] = None
        self.__core_memory_storage_package_list: Optional[list[str]] = None
        self.__core_work_pattern_package_list: Optional[list[str]] = None

    @property
    def base_info_appname(self) -> Optional[str]:
        """Return the appname of the application."""
        return self.__base_info_appname

    @property
    def core_default_package_list(self) -> Optional[list[str]]:
        """Return the default package list of the core."""
        return self.__core_default_package_list

    @property
    def core_agent_package_list(self) -> Optional[list[str]]:
        """Return the agent package list of the core."""
        return self.__core_agent_package_list

    @property
    def core_knowledge_package_list(self) -> Optional[list[str]]:
        """Return the knowledge package list of the core."""
        return self.__core_knowledge_package_list

    @property
    def core_llm_package_list(self) -> Optional[list[str]]:
        """Return the llm package list of the core."""
        return self.__core_llm_package_list

    @property
    def core_planner_package_list(self) -> Optional[list[str]]:
        """Return the planner package list of the core."""
        return self.__core_planner_package_list

    @property
    def core_tool_package_list(self) -> Optional[list[str]]:
        """Return the tool package list of the core."""
        return self.__core_tool_package_list

    @property
    def core_memory_package_list(self) -> Optional[list[str]]:
        """Return the memory package list of the core."""
        return self.__core_memory_package_list

    @property
    def core_service_package_list(self) -> Optional[list[str]]:
        """Return the service package list of the core."""
        return self.__core_service_package_list

    @property
    def core_sqldb_wrapper_package_list(self) -> Optional[list[str]]:
        """Return the sql db wrapper package list of the core."""
        return self.__core_sqldb_wrapper_package_list

    @property
    def core_prompt_package_list(self) -> Optional[list[str]]:
        return self.__core_prompt_package_list

    @property
    def core_product_package_list(self) -> Optional[list[str]]:
        return self.__core_product_package_list

    @property
    def core_workflow_package_list(self) -> Optional[list[str]]:
        return self.__core_workflow_package_list

    @property
    def core_embedding_package_list(self) -> Optional[list[str]]:
        """Return the embedding package list of the core."""
        return self.__core_embedding_package_list

    @property
    def core_doc_processor_package_list(self) -> Optional[list[str]]:
        """Return the document processor package list of the core."""
        return self.__core_doc_processor_package_list

    @property
    def core_reader_package_list(self) -> Optional[list[str]]:
        """Return the reader package list of the core."""
        return self.__core_reader_package_list

    @property
    def core_store_package_list(self) -> Optional[list[str]]:
        """Return the store package list of the core."""
        return self.__core_store_package_list

    @property
    def core_rag_router_package_list(self) -> Optional[list[str]]:
        """Return the RAG router package list of the core."""
        return self.__core_rag_router_package_list

    @property
    def core_query_paraphraser_package_list(self) -> Optional[list[str]]:
        """Return the query paraphraser package list of the core."""
        return self.__core_query_paraphraser_package_list

    @property
    def core_memory_compressor_package_list(self) -> Optional[list[str]]:
        """Return the memory compressor package list of the core."""
        return self.__core_memory_compressor_package_list

    @property
    def core_memory_storage_package_list(self) -> Optional[list[str]]:
        """Return the memory storage package list of the core."""
        return self.__core_memory_storage_package_list

    @property
    def core_work_pattern_package_list(self) -> Optional[list[str]]:
        """Return the work pattern package list of the core."""
        return self.__core_work_pattern_package_list

    def load_by_configer(self, configer: Configer) -> 'AppConfiger':
        """Load the AppConfiger by the given Configer.

        Args:
            configer(Configer): the Configer object
        Returns:
            AppConfiger: the AppConfiger object
        """
        self.__configer = configer
        self.__base_info_appname = configer.value.get('BASE_INFO', {}).get('appname')
        self.__core_default_package_list = configer.value.get('CORE_PACKAGE', {}).get('default')
        self.__core_agent_package_list = configer.value.get('CORE_PACKAGE', {}).get('agent')
        self.__core_knowledge_package_list = configer.value.get('CORE_PACKAGE', {}).get('knowledge')
        self.__core_llm_package_list = configer.value.get('CORE_PACKAGE', {}).get('llm')
        self.__core_planner_package_list = configer.value.get('CORE_PACKAGE', {}).get('planner')
        self.__core_tool_package_list = configer.value.get('CORE_PACKAGE', {}).get('tool')
        self.__core_memory_package_list = configer.value.get('CORE_PACKAGE', {}).get('memory')
        self.__core_service_package_list = configer.value.get('CORE_PACKAGE', {}).get('service')
        self.__core_sqldb_wrapper_package_list = configer.value.get('CORE_PACKAGE', {}).get('sqldb_wrapper')
        self.__core_prompt_package_list = configer.value.get('CORE_PACKAGE', {}).get('prompt')
        self.__core_product_package_list = configer.value.get('CORE_PACKAGE', {}).get('product')
        self.__core_workflow_package_list = configer.value.get('CORE_PACKAGE', {}).get('workflow')
        self.__core_embedding_package_list = configer.value.get('CORE_PACKAGE', {}).get('embedding')
        self.__core_doc_processor_package_list = configer.value.get('CORE_PACKAGE', {}).get('doc_processor')
        self.__core_reader_package_list = configer.value.get('CORE_PACKAGE', {}).get('reader')
        self.__core_store_package_list = configer.value.get('CORE_PACKAGE', {}).get('store')
        self.__core_rag_router_package_list = configer.value.get('CORE_PACKAGE', {}).get('rag_router')
        self.__core_query_paraphraser_package_list = configer.value.get('CORE_PACKAGE', {}).get('query_paraphraser')
        self.__core_memory_compressor_package_list = configer.value.get('CORE_PACKAGE', {}).get('memory_compressor')
        self.__core_memory_storage_package_list = configer.value.get('CORE_PACKAGE', {}).get('memory_storage')
        self.__core_work_pattern_package_list = configer.value.get('CORE_PACKAGE', {}).get('work_pattern')
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/31 14:25
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: web_pdf_reader.py
from io import BytesIO
from typing import List
import requests

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class WebPdfReader(Reader):
    """The agentUniverse(aU) web pdf reader.

    The pdf file will be downloaded and then parsed by `pdfminer.six`.
    """

    def _load_data(self, web_pdf_url: str) -> List[Document]:
        if web_pdf_url is None:
            return []
        response = requests.get(web_pdf_url)
        if response.status_code == 200:
            # download the pdf file and convert it into a memory file.
            pdf_memory_file = BytesIO(response.content)
            try:
                from pdfminer.high_level import extract_text_to_fp
            except ImportError:
                raise ImportError(
                    "pdfminer.six is required to read PDF files: `pip install pdfminer.six`"
                )
            # parse the pdf file and get the text content.
            with BytesIO() as output_string:
                extract_text_to_fp(pdf_memory_file, output_string, output_type='text', codec='utf-8')
                text = output_string.getvalue().decode('utf-8')
                return [Document(text=text, metadata={"source": web_pdf_url})]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 11:41
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: env.py

import os


def get_from_env(env_key: str) -> str:
    if env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/26 10:10
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm.py
# llm model name list
LLM_MODEL_NAME = {
    'demo_llm': ['gpt-3.5-turbo', 'gpt-4o', 'gpt-4-turbo', 'gpt-4o-mini', 'gpt-4'],
    'default_openai_llm': ['gpt-3.5-turbo', 'gpt-4o', 'gpt-4-turbo', 'gpt-4o-mini', 'gpt-4'],
    'qwen_llm': ['qwen-max', 'qwen-long', 'qwen-turbo', 'qwen2-72b-instruct', 'qwen2-7b-instruct', 'qwen1.5-110b-chat',
                 'qwen1.5-72b-chat', 'qwen1.5-7b-chat', 'qwen2.5-72b-instruct', 'qwen-plus'],
    'default_qwen_llm': ['qwen-max', 'qwen-long', 'qwen-turbo', 'qwen2-72b-instruct', 'qwen2-7b-instruct',
                         'qwen1.5-110b-chat', 'qwen1.5-72b-chat', 'qwen1.5-7b-chat', 'qwen2.5-72b-instruct',
                         'qwen-plus'],
    'wenxin_llm': ['ERNIE-Speed-AppBuilder-8K-0516', 'ERNIE-Lite-8K-0725', 'ERNIE-Speed-128K', 'ERNIE-3.5-128K',
                   'ERNIE-3.5-8K-0701', 'ERNIE-4.0-8K-0613', 'ERNIE-4.0-8K-Preview', 'ERNIE-3.5-8K-Preview',
                   'ERNIE-Tiny-8K', 'ERNIE-4.0-8K-Latest', 'ERNIE-4.0-Turbo-8K'],
    'default_wenxin_llm': ['ERNIE-Speed-AppBuilder-8K-0516', 'ERNIE-Lite-8K-0725', 'ERNIE-Speed-128K', 'ERNIE-3.5-128K',
                           'ERNIE-3.5-8K-0701', 'ERNIE-4.0-8K-0613', 'ERNIE-4.0-8K-Preview', 'ERNIE-3.5-8K-Preview',
                           'ERNIE-Tiny-8K', 'ERNIE-4.0-8K-Latest', 'ERNIE-4.0-Turbo-8K'],
    'kimi_llm': ['moonshot-v1-8k', 'moonshot-v1-32k', 'moonshot-v1-128k'],
    'default_kimi_llm': ['moonshot-v1-8k', 'moonshot-v1-32k', 'moonshot-v1-128k'],
    'deep_seek_llm': ['deepseek-chat', 'deepseek-coder'],
    'default_deepseek_llm': ['deepseek-chat', 'deepseek-coder'],
    'baichuan_llm': ['Baichuan4', 'Baichuan3-Turbo', 'Baichuan3-Turbo-128k', 'Baichuan2-Turbo', 'Baichuan2-Turbo-192k'],
    'default_baichuan_llm': ['Baichuan4', 'Baichuan3-Turbo', 'Baichuan3-Turbo-128k', 'Baichuan2-Turbo',
                             'Baichuan2-Turbo-192k'],
}



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/22 16:01
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: wenxin_langchain_instance.py


from typing import List, Optional, Any, Dict, Iterator, AsyncIterator

from langchain_community.chat_models import QianfanChatEndpoint
from langchain_community.chat_models.baidu_qianfan_endpoint import _convert_dict_to_message
from langchain_core.callbacks import CallbackManagerForLLMRun, AsyncCallbackManagerForLLMRun
from langchain_core.messages import BaseMessage, AIMessage, AIMessageChunk
from langchain_core.outputs import ChatResult, ChatGeneration, ChatGenerationChunk

from agentuniverse.llm.llm import LLM


class WenXinLangChainInstance(QianfanChatEndpoint):
    llm: LLM = None

    def __init__(self, llm: LLM):
        init_params = {"qianfan_ak": llm.qianfan_ak, "qianfan_sk": llm.qianfan_sk, "model": llm.model_name,
                       "max_tokens": llm.max_tokens, "timeout": llm.request_timeout,
                       'max_retries': llm.max_retries if llm.max_retries else 2,
                       'streaming': llm.streaming if llm.streaming else False,
                       'temperature': llm.temperature if llm.temperature else 0.7, 'llm': llm}
        super().__init__(**init_params)

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> ChatResult:
        """Call out to an qianfan models endpoint for each generation with a prompt.
              Args:
                  messages: The messages to pass into the model.
                  stop: Optional list of stop words to use when generating.
              Returns:
                  The string generated by the model.

              Example:
                  .. code-block:: python
                      response = qianfan_model.invoke("Tell me a joke.")
              """
        if self.streaming:
            completion = ""
            token_usage = {}
            chat_generation_info: Dict = {}
            for chunk in self._stream(messages, stop, run_manager, **kwargs):
                chat_generation_info = (
                    chunk.generation_info
                    if chunk.generation_info is not None
                    else chat_generation_info
                )
                completion += chunk.text
            lc_msg = AIMessage(content=completion, additional_kwargs={})
            gen = ChatGeneration(
                message=lc_msg,
                generation_info=dict(finish_reason="stop"),
            )
            return ChatResult(
                generations=[gen],
                llm_output={
                    "token_usage": chat_generation_info.get("usage", {}),
                    "model_name": self.model,
                },
            )
        params = self._convert_prompt_msg_params(messages, **kwargs)
        params["stop"] = stop
        response_payload = self.llm.call(**params).raw
        lc_msg = _convert_dict_to_message(response_payload)
        gen = ChatGeneration(
            message=lc_msg,
            generation_info={
                "finish_reason": "stop",
                **response_payload.get("body", {}),
            },
        )
        token_usage = response_payload.get("usage", {})
        llm_output = {"token_usage": token_usage, "model_name": self.model}
        return ChatResult(generations=[gen], llm_output=llm_output)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.streaming:
            completion = ""
            token_usage = {}
            chat_generation_info: Dict = {}
            async for chunk in self._astream(messages, stop, run_manager, **kwargs):
                chat_generation_info = (
                    chunk.generation_info
                    if chunk.generation_info is not None
                    else chat_generation_info
                )
                completion += chunk.text

            lc_msg = AIMessage(content=completion, additional_kwargs={})
            gen = ChatGeneration(
                message=lc_msg,
                generation_info=dict(finish_reason="stop"),
            )
            return ChatResult(
                generations=[gen],
                llm_output={
                    "token_usage": chat_generation_info.get("usage", {}),
                    "model_name": self.model,
                },
            )
        params = self._convert_prompt_msg_params(messages, **kwargs)
        params["stop"] = stop
        response_payload = await self.llm.acall(**params)
        response_payload = response_payload.raw
        lc_msg = _convert_dict_to_message(response_payload)
        generations = []
        gen = ChatGeneration(
            message=lc_msg,
            generation_info={
                "finish_reason": "stop",
                **response_payload.get("body", {}),
            },
        )
        generations.append(gen)
        token_usage = response_payload.get("usage", {})
        llm_output = {"token_usage": token_usage, "model_name": self.model}
        return ChatResult(generations=generations, llm_output=llm_output)

    def _stream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        """Call out to an qianfan models endpoint for each generation with a prompt.
              Args:
                  messages: The messages to pass into the model.
                  stop: Optional list of stop words to use when generating.
              Returns:
                  The string generated by the model.

              Example:
                  .. code-block:: python
                      response = qianfan_model.invoke("Tell me a joke.")
              """
        params = self._convert_prompt_msg_params(messages, **kwargs)
        params["stop"] = stop
        params["stream"] = True
        for res in self.llm.call(**params):
            if res:
                res = res.raw
                msg = _convert_dict_to_message(res)
                additional_kwargs = msg.additional_kwargs.get("function_call", {})
                chunk = ChatGenerationChunk(
                    text=res["result"],
                    message=AIMessageChunk(
                        content=msg.content,
                        role="assistant",
                        additional_kwargs=additional_kwargs,
                    ),
                    generation_info=msg.additional_kwargs,
                )
                if run_manager:
                    run_manager.on_llm_new_token(chunk.text, chunk=chunk)
                yield chunk

    async def _astream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        params = self._convert_prompt_msg_params(messages, **kwargs)
        params["stop"] = stop
        params["stream"] = True
        async for res in await self.llm.acall(**params):
            res = res.raw
            if res:
                msg = _convert_dict_to_message(res)
                additional_kwargs = msg.additional_kwargs.get("function_call", {})
                chunk = ChatGenerationChunk(
                    text=res["result"],
                    message=AIMessageChunk(
                        content=msg.content,
                        role="assistant",
                        additional_kwargs=additional_kwargs,
                    ),
                    generation_info=msg.additional_kwargs,
                )
                if run_manager:
                    await run_manager.on_llm_new_token(chunk.text, chunk=chunk)
                yield chunk



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/12 19:22
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: prompt_model.py
"""Agent Prompt Model module."""
from typing import Optional

from pydantic import BaseModel

from agentuniverse.agent.memory.enum import ChatMessageEnum


class AgentPromptModel(BaseModel):
    """Agent Prompt Model class."""

    introduction: Optional[str] = None
    target: Optional[str] = None
    instruction: Optional[str] = None
    _message_type_mapping: dict[str, str] = {'introduction': ChatMessageEnum.SYSTEM.value,
                                             'target': ChatMessageEnum.SYSTEM.value,
                                             'instruction': ChatMessageEnum.HUMAN.value}

    def __add__(self, other):
        """Merge two objects into one object."""
        merged_object = AgentPromptModel()
        for key in set(self.__dict__.keys()).union(other.__dict__.keys()):
            value = getattr(self, key, None)
            if value is None:
                value = getattr(other, key, None)
            setattr(merged_object, key, value)
        return merged_object

    def __bool__(self):
        """ Check whether the object is empty.

        Return True if one of the introduction, target and instruction attribute is not empty.
        Return False otherwise.
        """
        return bool(self.introduction or self.target or self.instruction)

    def get_message_type(self, attribute_name: str) -> str:
        """ Get the message type of the attribute in the agent prompt model.

            Args:
                attribute_name (str): The name of the attribute.
                
            Returns:
                str: The message type of the attribute(system/human/ai).
        """
        return self._message_type_mapping.get(attribute_name, ChatMessageEnum.HUMAN.value)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/17 16:54
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: executing_agent_template.py
import asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed
import time
from typing import Optional

from langchain_core.output_parsers import StrOutputParser

from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.base.util.common_util import stream_output
from agentuniverse.base.util.logging.logging_util import LOGGER
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.prompt import Prompt


class ExecutingAgentTemplate(AgentTemplate):
    _context_values: Optional[dict] = {}

    class Config:
        arbitrary_types_allowed = True

    def input_keys(self) -> list[str]:
        return ['input', 'planning_result']

    def output_keys(self) -> list[str]:
        return ['executing_result']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        agent_input['framework'] = input_object.get_data('planning_result').get_data('framework')
        agent_input['expert_framework'] = input_object.get_data('expert_framework', {}).get('executing')
        return agent_input

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM, prompt: Prompt,
                           **kwargs) -> dict:
        return self._execute_tasks(input_object, agent_input, memory, llm, prompt)

    async def customized_async_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM,
                                       prompt: Prompt, **kwargs) -> dict:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._execute_tasks, input_object, agent_input, memory,
                                          llm, prompt)

    def _execute_tasks(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM,
                       prompt: Prompt, **kwargs) -> dict:
        self._context_values: dict = FrameworkContextManager().get_all_contexts()

        framework = agent_input.get('framework', [])

        with ThreadPoolExecutor(max_workers=min(len(framework), 10),
                                thread_name_prefix="executing_agent_template") as thread_executor:
            futures = []
            for i, subtask in enumerate(framework):
                future = thread_executor.submit(self._execute_subtask, subtask, input_object, agent_input, i, memory,
                                                llm, prompt)
                futures.append(future)
                time.sleep(1)

            executing_result = [future.result() for future in as_completed(futures)]

        executing_result.sort(key=lambda x: x['index'])
        return {'executing_result': [result for result in executing_result],
                'output_stream': input_object.get_data('output_stream', None)}

    def _execute_subtask(self, subtask, input_object, agent_input, index, memory, llm, prompt) -> dict:
        context_tokens = {}
        try:
            # pass the framework context into the thread.
            for var_name, var_value in self._context_values.items():
                token = FrameworkContextManager().set_context(var_name, var_value)
                context_tokens[var_name] = token

            input_object_copy = InputObject(input_object.to_dict())
            agent_input_copy = dict(agent_input)

            self._process_tool_inputs(input_object_copy, subtask)

            knowledge_res = self.invoke_knowledge(subtask, input_object_copy)
            tools_res = self.invoke_tools(input_object_copy)
            agent_input_copy['background'] = f"knowledge result: {knowledge_res} \n\n tools result: {tools_res}"
            agent_input_copy['input'] = subtask

            process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input_copy)
            assemble_memory_input(memory, agent_input_copy)

            chain = prompt.as_langchain() | llm.as_langchain_runnable(
                self.agent_model.llm_params()) | StrOutputParser()
            res = self.invoke_chain(chain, agent_input_copy, input_object_copy)

            assemble_memory_output(memory=memory,
                                   agent_input=agent_input,
                                   content=f"Human: {agent_input.get('input')}, AI: {res}")
            return {
                'index': index,
                'input': f"Question {index + 1}: {subtask}",
                'output': f"Answer {index + 1}: {res}"
            }
        finally:
            # clear the framework context.
            for var_name, token in context_tokens.items():
                FrameworkContextManager().reset_context(var_name, token)

    def parse_result(self, agent_result: dict) -> dict:
        # add executing agent final result into the stream output.
        stream_output(agent_result.pop('output_stream'),
                      {"data": {
                          'output': agent_result.get('executing_result'),
                          "agent_info": self.agent_model.info
                      }, "type": "executing"})

        # add executing agent log info.
        logger_info = f"\nExecuting agent execution result is :\n"
        if agent_result.get('executing_result'):
            for index, one_exec_res in enumerate(agent_result.get('executing_result')):
                one_exec_log_info = f"[{index + 1}] input: {one_exec_res['input']}\n"
                one_exec_log_info += f"[{index + 1}] output: {one_exec_res['output']}\n"
                logger_info += one_exec_log_info
        LOGGER.info(logger_info)

        return agent_result

    def _process_tool_inputs(self, input_object: InputObject, subtask: str) -> None:
        if not self.tool_names:
            return
        for tool_name in self.tool_names:
            tool = ToolManager().get_instance_obj(tool_name)
            if tool is not None:
                # note: only insert the first key of tool input.
                input_object.add_data(tool.input_keys[0], subtask)

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'ExecutingAgentTemplate':
        super().initialize_by_component_configer(component_configer)
        self.prompt_version = self.agent_model.profile.get('prompt_version', 'default_executing_agent.cn')
        self.validate_required_params()
        return self

    def validate_required_params(self):
        if not self.llm_name:
            raise ValueError(f'llm_name of the agent {self.agent_model.info.get("name")}'
                             f' is not set, please go to the agent profile configuration'
                             ' and set the `name` attribute in the `llm_model`.')



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 17:49
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: claude_llm.py

from typing import Optional, Any, Union, Iterator, AsyncIterator

from pydantic import Field

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.openai_style_llm import OpenAIStyleLLM

DEEpSEEkMAXCONTETNLENGTH = {
    "deepseek-chat": 32000,
    "deepseek-coder": 32000,
    "claude-3-haiku-20240307": 200000,
    "claude-2.1": 200000,
    "claude-2.0": 100000,
    "claude-instant-1.2": 100000
}


class DefaultDeepSeekLLM(OpenAIStyleLLM):
    """The agentUniverse default openai llm module.

    LLM parameters, such as name/description/model_name/max_tokens,
    are injected into this class by the default_openai_llm.yaml configuration.
    """

    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("DEEPSEEK_API_KEY"))
    organization: Optional[str] = Field(default_factory=lambda: get_from_env("DEEPSEEK_ORGANIZATION"))
    api_base: Optional[str] = Field(default_factory=lambda: get_from_env("DEEPSEEK_API_BASE"))
    proxy: Optional[str] = Field(default_factory=lambda: get_from_env("DEEPSEEK_PROXY"))

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """ The call method of the LLM.

        Users can customize how the model interacts by overriding call method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return super()._call(messages, **kwargs)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """ The async call method of the LLM.

        Users can customize how the model interacts by overriding acall method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return await super()._acall(messages, **kwargs)

    def max_context_length(self) -> int:
        """Max context length.

          The total length of input tokens and generated tokens is limited by the openai model's context length.
          """
        return DEEpSEEkMAXCONTETNLENGTH.get(self.model_name, 4096)




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 15:57
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_configer.py
from typing import Optional
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer


class WorkflowConfiger(ComponentConfiger):
    """The WorkflowConfiger class, which is used to load and manage the workflow configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the WorkflowConfiger."""
        super().__init__(configer)
        self.__id: Optional[str] = None
        self.__name: Optional[str] = None
        self.__description: Optional[str] = None
        self.__graph: Optional[dict] = None

    @property
    def id(self) -> Optional[str]:
        """Return the id of the Workflow."""
        return self.__id

    @property
    def name(self) -> Optional[str]:
        """Return the name of the Workflow."""
        return self.__name

    @property
    def description(self) -> Optional[str]:
        """Return the description of the Workflow."""
        return self.__description

    @property
    def graph(self) -> Optional[dict]:
        return self.__graph

    def load(self) -> 'WorkflowConfiger':
        """Load the configuration by the Configer object.
        Returns:
            WorkflowConfiger: the WorkflowConfiger object
        """
        return self.load_by_configer(self.__configer)

    def load_by_configer(self, configer: Configer) -> 'WorkflowConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            WorkflowConfiger: the WorkflowConfiger object
        """
        super().load_by_configer(configer)

        try:
            self.__id = configer.value.get('id')
            self.__name = configer.value.get('name')
            self.__description = configer.value.get('description')
            self.__graph = configer.value.get('graph')
        except Exception as e:
            raise Exception(f"Failed to parse the Workflow configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 20:02
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm_node.py
import re
from typing import List, Optional

from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import NodeData, Node
from agentuniverse.workflow.node.node_config import LLMNodeInputParams, NodeOutputParams
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.workflow_output import WorkflowOutput


class LLMNodeData(NodeData):
    inputs: Optional[LLMNodeInputParams] = None


class LLMNode(Node):
    """The basic class of the llm node."""

    _data_cls = LLMNodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.LLM

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        inputs: LLMNodeInputParams = self._data.inputs

        param_map = {
            'model_name': None,
            'temperature': None,
            'prompt': None,
            'id': None
        }

        for llm_param in inputs.llm_param:
            if llm_param.name in param_map:
                if isinstance(llm_param.value, str):
                    param_map[llm_param.name] = llm_param.value
                else:
                    param_map[llm_param.name] = llm_param.value.get('content', None)

        model_name = param_map['model_name']
        temperature = param_map['temperature']
        prompt = param_map['prompt']
        llm_id = param_map['id']

        llm: LLM = LLMManager().get_instance_obj(llm_id)
        if llm is None:
            raise ValueError("No llm with id {} was found.".format(llm_id))

        input_variables = re.findall(r'\{\{(.*?)\}\}', prompt)
        llm.set_by_agent_model(model_name=model_name, temperature=temperature)

        llm_input_params = self._resolve_input_params(inputs.input_param, workflow_output)
        for variable in input_variables:
            if variable not in llm_input_params:
                raise ValueError(f"The variable {variable} is not found in the input params.")
            prompt = prompt.replace('{{' + variable + '}}', llm_input_params[variable])

        messages = [{'role': 'user', 'content': prompt}]
        llm_output: LLMOutput = llm.call(messages=messages)

        # handle output parameters
        output_params: List[NodeOutputParams] = self._data.outputs
        output_params[0].value = llm_output.text
        workflow_output.workflow_parameters[self.id] = output_params

        return NodeOutput(node_id=self.id, status=NodeStatusEnum.SUCCEEDED, result=output_params)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 13:52
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: wenxin_llm.py
from typing import Any, Union, AsyncIterator, Iterator

import qianfan
from langchain_core.language_models import BaseLanguageModel
from pydantic import Field
from qianfan import QfResponse
from qianfan.resources.tools import tokenizer

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.wenxin_langchain_instance import WenXinLangChainInstance

TokenModelList = [
    'Ernie-4.0-8k',
    'ernie-3.5-8k',
    'ernie-speed-8k',
    'ernie-speed-128k',
    'ernie-lite-8k',
    'ernie-tiny-8k',
    'ernie-char-8k',
]


class WenXinLLM(LLM):
    """WenXin LLM wrapper for LangChain.
        This is a wrapper around the Qianfan Chat Completion API.
        It uses the QianfanChatEndpoint class from the langchain_community package.
        It also supports streaming and token counting.
        It can be used with any other LangChain tools or agents.

        Attributes:
            qianfan_ak (str): The access key of the Qianfan API.
                Defaults to the value of the QIANFAN_AK environment variable.
            qianfan_sk (str): The secret key of the Qianfan API.
                Defaults to the value of the QIANFAN_SK environment variable.
    """
    qianfan_ak: str = Field(default_factory=lambda: get_from_env("QIANFAN_AK"))
    qianfan_sk: str = Field(default_factory=lambda: get_from_env("QIANFAN_SK"))

    def _new_client(self):
        if self.client is None:
            self.client = qianfan.ChatCompletion(ak=self.qianfan_ak, sk=self.qianfan_sk)
        """Create a new Qianfan client."""
        return self.client

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """Run the OpenAI LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        self.client = self._new_client()
        client = self.client
        chat_completion = client.do(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            return self.parse_result(chat_completion)
        return self.generate_stream_result(chat_completion)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """Asynchronously run the OpenAI LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        self.async_client = self._new_client()
        async_client = self.async_client
        chat_completion = await async_client.ado(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            return self.parse_result(chat_completion)
        return self.agenerate_stream_result(chat_completion)

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        res = self._new_client().get_model_info(self.model_name)
        if res.max_input_tokens:
            return res.max_input_tokens
        return res.max_input_chars

    def get_num_tokens(self, text: str) -> int:
        model_name = ''
        if self.model_name.lower() in TokenModelList:
            model_name = self.model_name.lower()
        token_cnt = tokenizer.Tokenizer().count_tokens(
            text=text,
            mode='remote',
            model=model_name
        )
        return token_cnt

    @staticmethod
    def parse_result(chunk: QfResponse):
        text = chunk.body.get('result')
        if not text:
            return None
        return LLMOutput(text=text, raw=chunk)

    def generate_stream_result(self, chat_completion) -> Iterator[LLMOutput]:
        for chunk in chat_completion:
            data = self.parse_result(chunk)
            if data:
                yield data

    async def agenerate_stream_result(self, chat_completion: AsyncIterator) -> AsyncIterator[LLMOutput]:
        async for chunk in chat_completion:
            data = self.parse_result(chunk)
            if data:
                yield data

    def as_langchain(self) -> BaseLanguageModel:
        """Return an instance of the LangChain `BaseLanguageModel` class."""
        return WenXinLangChainInstance(llm=self)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/24 16:35
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: product_configer.py
from typing import Optional

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer

PRODUCT_COMPONENT_TYPE = ['AGENT', 'KNOWLEDGE', 'TOOL', 'PLANNER', 'LLM', 'PLUGIN']


class ProductConfiger(ComponentConfiger):
    """The ProductConfiger class, used to load and manage the product configuration."""

    _ComponentConfiger__metadata_class: Optional[str] = None
    _ComponentConfiger__metadata_module: Optional[str] = None

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the ProductConfiger."""
        super().__init__(configer)
        self.__nickname: Optional[str] = None
        self.__id: Optional[str] = None
        self.__type: Optional[str] = None
        self.__avatar: Optional[str] = None
        self.__description: Optional[str] = None
        self.__set_default_meta_info()

    @property
    def nickname(self) -> Optional[str]:
        """Nickname field."""
        return self.__nickname

    @property
    def id(self) -> Optional[str]:
        """ID field."""
        return self.__id

    @property
    def type(self) -> Optional[str]:
        """Type field."""
        return self.__type

    @property
    def avatar(self) -> Optional[str]:
        """Avatar field."""
        return self.__avatar
    
    @property
    def description(self) -> Optional[str]:
        """Description field."""
        return self.__description

    def __set_default_meta_info(self):
        """Set default instantiated class of product."""
        if (not hasattr(self, '_ComponentConfiger__metadata_module')
                or self._ComponentConfiger__metadata_module is None):
            self._ComponentConfiger__metadata_module = ("agentuniverse_product."
                                                        "base.product")
        if (not hasattr(self, '_ComponentConfiger__metadata_class')
                or self._ComponentConfiger__metadata_class is None):
            self._ComponentConfiger__metadata_class = 'Product'

    def load(self) -> 'ProductConfiger':
        """Setting property using own configer member property.

        Returns:
            ProductConfiger: A ProductConfiger instance.
        """
        return self.load_by_configer(self.configer)

    def load_by_configer(self, configer: Configer) -> 'ProductConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            ProductConfiger: the ProductConfiger object
        """
        super().load_by_configer(configer)
        self.__set_default_meta_info()
        try:
            self.__nickname = configer.value.get('nickname')
            if self.__nickname is None:
                raise ValueError(f"Product nickname is required parameter.")
            self.__id = configer.value.get('id')
            if self.__id is None:
                raise ValueError(f"Product id is required parameter.")
            self.__type = configer.value.get('type')
            self.__avatar = configer.value.get('avatar')
            self.__description = configer.value.get('description')
            if self.__type not in PRODUCT_COMPONENT_TYPE:
                raise ValueError(f"Invalid product type: {self.__type}")
        except Exception as e:
            raise Exception(f"Failed to parse the Product configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 16:20
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: default_openai_llm.py
from typing import Any, Optional, Iterator, Union, AsyncIterator

from pydantic import Field

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.openai_style_llm import OpenAIStyleLLM

OPENAI_MAX_CONTEXT_LENGTH = {
    "gpt-3.5-turbo": 4096,
    "gpt-3.5-turbo-0301": 4096,
    "gpt-3.5-turbo-0613": 4096,
    "gpt-3.5-turbo-16k": 16384,
    "gpt-3.5-turbo-16k-0613": 16384,
    "gpt-35-turbo": 4096,
    "gpt-35-turbo-16k": 16384,
    "gpt-3.5-turbo-1106": 16384,
    "gpt-3.5-turbo-0125": 16384,
    "gpt-4-0314": 8192,
    "gpt-4": 8192,
    "gpt-4-32k": 32768,
    "gpt-4-32k-0613": 32768,
    "gpt-4-0613": 8192,
    "gpt-4-1106-preview": 128000,
    "gpt-4-turbo": 128000,
    "gpt-4o": 128000,
    "gpt-4o-2024-05-13": 128000,
}


class DefaultOpenAILLM(OpenAIStyleLLM):
    """The agentUniverse default openai llm module.

    LLM parameters, such as name/description/model_name/max_tokens,
    are injected into this class by the default_openai_llm.yaml configuration.
    """
    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_API_KEY"))
    organization: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_ORGANIZATION"))
    api_base: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_API_BASE"))
    proxy: Optional[str] = Field(default_factory=lambda: get_from_env("OPENAI_PROXY"))

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """ The call method of the LLM.

        Users can customize how the model interacts by overriding call method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return super()._call(messages, **kwargs)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """ The async call method of the LLM.

        Users can customize how the model interacts by overriding acall method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return await super()._acall(messages, **kwargs)

    def max_context_length(self) -> int:
        """Max context length.

          The total length of input tokens and generated tokens is limited by the openai model's context length.
          """
        return OPENAI_MAX_CONTEXT_LENGTH.get(self.model_name, 4096)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/25 16:04
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: service_configer.py

from typing import Optional

from ..agent.agent import Agent
from ..agent.agent_manager import AgentManager
from ..base.config.component_configer.component_configer import ComponentConfiger
from ..base.config.configer import Configer


class ServiceConfiger(ComponentConfiger):
    """The ServiceConfiger class, used to load and manage the service
    configuration."""

    _ComponentConfiger__metadata_class: Optional[str] = None
    _ComponentConfiger__metadata_module: Optional[str] = None

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the ServiceConfiger."""
        super().__init__(configer)
        self.__name: Optional[str] = None
        self.__description: Optional[str] = None
        self.__agent: Optional[Agent] = None
        self.__set_default_meta_info()

    @property
    def name(self) -> Optional[str]:
        """Name field."""
        return self.__name

    @property
    def description(self) -> Optional[str]:
        """Description field."""
        return self.__description

    @property
    def agent(self) -> Optional[Agent]:
        """Agent field."""
        return self.__agent

    def __set_default_meta_info(self):
        """Set default instantiated class of service."""
        if (not hasattr(self, '_ComponentConfiger__metadata_module')
                or self._ComponentConfiger__metadata_module is None):
            self._ComponentConfiger__metadata_module = ("agentuniverse."
                                                        "agent_serve.service")
        if (not hasattr(self, '_ComponentConfiger__metadata_class')
                or self._ComponentConfiger__metadata_class is None):
            self._ComponentConfiger__metadata_class = 'Service'

    def load(self) -> 'ServiceConfiger':
        """Setting property using own configer member property.

        Returns:
            ServiceConfiger: A ServiceConfiger instance.
        """
        return self.load_by_configer(self.configer)

    def load_by_configer(self, configer: Configer) -> 'ServiceConfiger':
        """Initialize self using given configer, get ServiceConfiger property
        from it.
        Args:
            configer(Configer): A Configer instance.
        Returns:
            ServiceConfiger: A ServiceConfiger instance.
        """
        super().load_by_configer(configer)
        agent_code = configer.value.get('agent')
        self.__set_default_meta_info()
        try:
            self.__name = configer.value.get('name')
            self.__description = configer.value.get('description')
            agent_manager: AgentManager = AgentManager()
            self.__agent = agent_manager.get_instance_obj(agent_code)
            if not self.__agent:
                raise ValueError
        except ValueError:
            raise ValueError(f"No such Agent: {agent_code}")
        except Exception as e:
            raise Exception(f"Failed to parse the Agent configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/15 10:05
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory.py
from typing import Optional, List
from langchain_core.memory import BaseMemory
from pydantic import Extra

from agentuniverse.agent.memory.enum import MemoryTypeEnum
from agentuniverse.agent.memory.memory_compressor.memory_compressor import MemoryCompressor
from agentuniverse.agent.memory.memory_compressor.memory_compressor_manager import MemoryCompressorManager
from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.agent.memory.memory_storage.memory_storage_manager import MemoryStorageManager
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.memory_configer import MemoryConfiger
from agentuniverse.base.util.memory_util import get_memory_tokens


class Memory(ComponentBase):
    """The basic class for memory model.

    Attributes:
        name (Optional[str]): The name of the memory class.
        description (Optional[str]): The description of the memory class.
        type (MemoryTypeEnum): The type of the memory class including `long-term` and `short-term`.
        memory_key (Optional[str]): The name of the memory key in the prompt.
        max_tokens (int): The maximum number of tokens allowed in the prompt.
        memory_compressor (Optional[str]): The name of the memory compressor instance.
        memory_storages (Optional[str]): The name list of the memory storage instances.
        memory_retrieval_storage (Optional[str]): The name of the memory retrieval storage instance.
    """

    name: Optional[str] = ""
    description: Optional[str] = None
    type: MemoryTypeEnum = None
    memory_key: Optional[str] = 'chat_history'
    max_tokens: int = 2000
    memory_compressor: Optional[str] = None
    memory_storages: Optional[List[str]] = ['local_memory_storage']
    memory_retrieval_storage: Optional[str] = None

    class Config:
        extra = Extra.allow

    def __init__(self, **kwargs):
        super().__init__(component_type=ComponentEnum.MEMORY, **kwargs)

    def as_langchain(self) -> BaseMemory:
        """Convert the agentUniverse(aU) memory class to the langchain memory class."""
        pass

    def add(self, message_list: List[Message], session_id: str = None, agent_id: str = None,
            **kwargs) -> None:
        """Add messages to the memory."""
        if not message_list:
            return
        for storage in self.memory_storages:
            memory_storage: MemoryStorage = MemoryStorageManager().get_instance_obj(storage)
            if memory_storage:
                memory_storage.add(message_list, session_id, agent_id, **kwargs)

    def delete(self, session_id: str = None, **kwargs) -> None:
        """Delete messages from the memory."""
        for storage in self.memory_storages:
            memory_storage: MemoryStorage = MemoryStorageManager().get_instance_obj(storage)
            if memory_storage:
                memory_storage.delete(session_id, **kwargs)

    def get(self, session_id: str = None, agent_id: str = None, **kwargs) -> List[Message]:
        """Get messages from the memory."""
        memory_storage: MemoryStorage = MemoryStorageManager().get_instance_obj(self.memory_retrieval_storage)
        if memory_storage:
            memories = memory_storage.get(session_id, agent_id, **kwargs)
            return self.prune(memories)
        return []

    def prune(self, memories: List[Message]) -> List[Message]:
        if not memories:
            return []
        new_memories = memories[:]

        agent_llm_name = self.agent_llm_name if hasattr(self, 'agent_llm_name') else None
        tokens = get_memory_tokens(new_memories, agent_llm_name)

        if tokens <= self.max_tokens:
            return new_memories

        pruned_memories = []
        while tokens > self.max_tokens:
            pruned_memory = new_memories.pop(0)
            pruned_memories.append(pruned_memory)
            tokens = get_memory_tokens(new_memories, agent_llm_name)

        if pruned_memories:
            memory_compressor: MemoryCompressor = MemoryCompressorManager().get_instance_obj(self.memory_compressor)
            if memory_compressor:
                compressed_memory = memory_compressor.compress_memory(pruned_memories, self.max_tokens - tokens)
                if compressed_memory:
                    new_memories.insert(0, Message(content=compressed_memory))
        return new_memories

    def set_by_agent_model(self, **kwargs):
        """ Assign values of parameters to the Memory model in the agent configuration."""
        # note: default shallow copy
        copied_obj = self.model_copy()
        if 'memory_key' in kwargs and kwargs['memory_key']:
            copied_obj.memory_key = kwargs['memory_key']
        if 'max_tokens' in kwargs and kwargs['max_tokens']:
            copied_obj.max_tokens = kwargs['max_tokens']
        if 'agent_llm_name' in kwargs and kwargs['agent_llm_name']:
            copied_obj.agent_llm_name = kwargs['agent_llm_name']
        return copied_obj

    def get_instance_code(self) -> str:
        """Return the full name of the memory."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f'{appname}.{self.component_type.value.lower()}.{self.name}'

    def initialize_by_component_configer(self, component_configer: MemoryConfiger) -> 'Memory':
        """Initialize the memory by the ComponentConfiger object.
        Args:
            component_configer(MemoryConfiger): the ComponentConfiger object
        Returns:
            Memory: the Memory object
        """
        if component_configer.name:
            self.name = component_configer.name
        if component_configer.description:
            self.description = component_configer.description
        if component_configer.type:
            self.type = next((member for member in MemoryTypeEnum if member.value == component_configer.type))
        if component_configer.memory_key:
            self.memory_key = component_configer.memory_key
        if component_configer.max_tokens:
            self.max_tokens = component_configer.max_tokens
        if component_configer.memory_compressor:
            self.memory_compressor = component_configer.memory_compressor
        if component_configer.memory_storages:
            self.memory_storages = component_configer.memory_storages
        if component_configer.memory_retrieval_storage:
            self.memory_retrieval_storage = component_configer.memory_retrieval_storage
        if not self.memory_retrieval_storage:
            self.memory_retrieval_storage = self.memory_storages[0]
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/31 21:22
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: react_planner.py
from typing import Sequence, Optional, Union, List

from langchain.agents.format_scratchpad import format_log_to_str
from langchain.agents.output_parsers import ReActSingleInputOutputParser

from langchain.agents import AgentExecutor, AgentOutputParser
from langchain.tools import Tool as LangchainTool
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import BasePromptTemplate
from langchain_core.runnables import RunnableConfig, RunnablePassthrough, Runnable
from langchain_core.tools import BaseTool, ToolsRenderer, render_text_description

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent import Agent
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.agent.plan.planner.react_planner.stream_callback import StreamOutPutCallbackHandler
from agentuniverse.base.util.agent_util import assemble_memory_input
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.chat_prompt import ChatPrompt
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel


class ReActPlanner(Planner):
    """ReAct planner class."""

    tools: list[LangchainTool] = None

    def invoke(self, agent_model: AgentModel, planner_input: dict,
               input_object: InputObject) -> dict:
        """Invoke the planner.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): The input parameters passed by the user.
        Returns:
            dict: The planner result.
        """

        memory: Memory = self.handle_memory(agent_model, planner_input)

        llm: LLM = self.handle_llm(agent_model)
        tools = self.acquire_tools(agent_model.action)
        prompt: Prompt = self.handle_prompt(agent_model, planner_input)
        process_llm_token(llm, prompt.as_langchain(), agent_model.profile, planner_input)
        assemble_memory_input(memory, planner_input)
        stop_sequence = []
        if agent_model.plan.get('stop_sequence'):
            stop_sequence = agent_model.profile.get('stop_sequence')
        agent = create_react_agent(llm.as_langchain(), tools, prompt.as_langchain(), stop_sequence=stop_sequence,
                                   bind_params=agent_model.llm_params())
        agent_executor = AgentExecutor(agent=agent, tools=tools,
                                       verbose=True,
                                       handle_parsing_errors=True,
                                       max_iterations=agent_model.plan.get('planner').get("max_iterations", 15))

        return agent_executor.invoke(input=planner_input, memory=memory.as_langchain() if memory else None,
                                     chat_history=planner_input.get(memory.memory_key) if memory else '',
                                     config=self.get_run_config(agent_model, input_object))

    @staticmethod
    def get_run_config(agent_model: AgentModel, input_object: InputObject) -> RunnableConfig:
        config = RunnableConfig()
        callbacks = []
        output_stream = input_object.get_data('output_stream')
        callbacks.append(StreamOutPutCallbackHandler(output_stream, agent_info=agent_model.info))
        config.setdefault("callbacks", callbacks)
        return config

    @staticmethod
    def acquire_tools(action) -> list[LangchainTool]:
        tool_names: list = action.get('tool') or list()
        lc_tools: list[LangchainTool] = list()
        for tool_name in tool_names:
            tool: Tool = ToolManager().get_instance_obj(tool_name)
            lc_tools.append(tool.as_langchain())
        knowledge: list = action.get('knowledge') or list()
        for knowledge_name in knowledge:
            knowledge_tool: Knowledge = KnowledgeManager().get_instance_obj(knowledge_name)
            lc_tools.append(knowledge_tool.as_langchain_tool())

        agents: list = action.get('agent') or list()
        for agent_name in agents:
            agent_tool: Agent = AgentManager().get_instance_obj(agent_name)
            lc_tools.append(agent_tool.as_langchain_tool())
        return lc_tools

    def handle_prompt(self, agent_model: AgentModel, planner_input: dict) -> Prompt:
        """Prompt module processing.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
        Returns:
            ChatPrompt: The chat prompt instance.
        """
        lc_tools_str: str = ''
        tools = self.acquire_tools(agent_model.action)
        for lc_tool in self.acquire_tools(agent_model.action):
            lc_tools_str += "tool name:" + lc_tool.name + " " + "tool description:" + lc_tool.description + '\n'
        lc_tool_names = "|".join([lc_tool.name for lc_tool in tools])
        planner_input['tool_names'] = lc_tool_names
        planner_input['tools'] = lc_tools_str
        planner_input['agent_scratchpad'] = ''

        profile: dict = agent_model.profile

        profile_prompt_model: AgentPromptModel = AgentPromptModel(introduction=profile.get('introduction'),
                                                                  target=profile.get('target'),
                                                                  instruction=profile.get('instruction'))

        # get the prompt by the prompt version
        prompt_version: str = profile.get('prompt_version')
        version_prompt: Prompt = PromptManager().get_instance_obj(prompt_version)

        if version_prompt is None and not profile_prompt_model:
            raise Exception("Either the `prompt_version` or `introduction & target & instruction`"
                            " in agent profile configuration should be provided.")
        if version_prompt:
            version_prompt_model: AgentPromptModel = AgentPromptModel(
                introduction=getattr(version_prompt, 'introduction', ''),
                target=getattr(version_prompt, 'target', ''),
                instruction=getattr(version_prompt, 'instruction', ''))
            profile_prompt_model = profile_prompt_model + version_prompt_model

        prompt = Prompt().build_prompt(profile_prompt_model, self.prompt_assemble_order)
        return prompt


def create_react_agent(
        llm: BaseLanguageModel,
        tools: Sequence[BaseTool],
        prompt: BasePromptTemplate,
        output_parser: Optional[AgentOutputParser] = None,
        tools_renderer: ToolsRenderer = render_text_description,
        *,
        stop_sequence: Union[bool, List[str]] = True,
        bind_params: Optional[dict],
) -> Runnable:
    missing_vars = {"tools", "tool_names", "agent_scratchpad"}.difference(
        prompt.input_variables + list(prompt.partial_variables)
    )
    if missing_vars:
        raise ValueError(f"Prompt missing required variables: {missing_vars}")

    prompt = prompt.partial(
        tools=tools_renderer(list(tools)),
        tool_names=", ".join([t.name for t in tools]),
    )
    if stop_sequence:
        stop = ["\nObservation"] if stop_sequence is True else stop_sequence
        llm_with_stop = llm.bind(stop=stop, **(bind_params or {}))
    else:
        llm_with_stop = llm.bind(**(bind_params or {}))
    output_parser = output_parser or ReActSingleInputOutputParser()
    agent = (
            RunnablePassthrough.assign(
                agent_scratchpad=lambda x: format_log_to_str(x["intermediate_steps"]),
            )
            | prompt
            | llm_with_stop
            | output_parser
    )
    return agent



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 23:17
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool_dto.py
from typing import Optional, List

from pydantic import BaseModel, Field


class ToolDTO(BaseModel):
    id: str = Field(description="ID")
    nickname: Optional[str] = Field(description="tool nickname", default="")
    avatar: Optional[str] = Field(description="tool avatar path", default="")
    description: Optional[str] = Field(description="tool description", default="")
    parameters: Optional[List[str]] = Field(description="tool parameters", default=[])
    openapi_schema: Optional[dict] = Field(description="openapi schema for tool", default={})
